#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import gzip
import ssl
import warnings
from unittest.mock import Mock, patch

import pytest
import requests
from requests.auth import HTTPBasicAuth

from elastic_transport import NodeConfig, RequestsHttpNode
from elastic_transport._node._base import DEFAULT_USER_AGENT


class TestRequestsHttpNode:
    def _get_mock_node(self, node_config, response_body=b"{}"):
        node = RequestsHttpNode(node_config)

        def _dummy_send(*args, **kwargs):
            dummy_response = Mock()
            dummy_response.headers = {}
            dummy_response.status_code = 200
            dummy_response.content = response_body
            dummy_response.request = args[0]
            dummy_response.cookies = {}
            _dummy_send.call_args = (args, kwargs)
            return dummy_response

        node.session.send = _dummy_send
        return node

    def _get_request(self, node, *args, **kwargs) -> requests.PreparedRequest:
        resp, data = node.perform_request(*args, **kwargs)
        status = resp.status
        assert 200 == status
        assert b"{}" == data

        timeout = kwargs.pop("request_timeout", node.config.request_timeout)
        args, kwargs = node.session.send.call_args
        assert timeout == kwargs["timeout"]
        assert 1 == len(args)
        return args[0]

    def test_close_session(self):
        node = RequestsHttpNode(NodeConfig("http", "localhost", 80))
        with patch.object(node.session, "close") as pool_close:
            node.close()
        pool_close.assert_called_with()

    def test_ssl_context(self):
        ctx = ssl.create_default_context()
        node = RequestsHttpNode(NodeConfig("https", "localhost", 80, ssl_context=ctx))
        adapter = node.session.get_adapter("https://localhost:80")
        assert adapter.poolmanager.connection_pool_kw["ssl_context"] is ctx

    def test_merge_headers(self):
        node = self._get_mock_node(
            NodeConfig("http", "localhost", 80, headers={"h1": "v1", "h2": "v2"})
        )
        req = self._get_request(node, "GET", "/", headers={"h2": "v2p", "h3": "v3"})
        assert req.headers["h1"] == "v1"
        assert req.headers["h2"] == "v2p"
        assert req.headers["h3"] == "v3"

    def test_default_headers(self):
        node = self._get_mock_node(NodeConfig("http", "localhost", 80))
        req = self._get_request(node, "GET", "/")
        assert req.headers == {
            "connection": "keep-alive",
            "user-agent": DEFAULT_USER_AGENT,
        }

    def test_no_http_compression(self):
        node = self._get_mock_node(
            NodeConfig("http", "localhost", 80, http_compress=False)
        )
        assert not node.config.http_compress
        assert "accept-encoding" not in node.headers

        node.perform_request("GET", "/")
        (req,), _ = node.session.send.call_args

        assert req.body is None
        assert "accept-encoding" not in req.headers
        assert "content-encoding" not in req.headers

        node.perform_request("GET", "/", body=b"hello, world!")
        (req,), _ = node.session.send.call_args

        assert req.body == b"hello, world!"
        assert "accept-encoding" not in req.headers
        assert "content-encoding" not in req.headers

    @pytest.mark.parametrize("empty_body", [None, b""])
    def test_http_compression(self, empty_body):
        node = self._get_mock_node(
            NodeConfig("http", "localhost", 80, http_compress=True)
        )
        assert node.config.http_compress is True
        assert node.headers["accept-encoding"] == "gzip"

        # 'content-encoding' shouldn't be set at a connection level.
        # Should be applied only if the request is sent with a body.
        assert "content-encoding" not in node.headers

        node.perform_request("GET", "/", body=b"{}")
        (req,), _ = node.session.send.call_args

        assert gzip.decompress(req.body) == b"{}"
        assert req.headers["accept-encoding"] == "gzip"
        assert req.headers["content-encoding"] == "gzip"

        node.perform_request("GET", "/", body=empty_body)
        (req,), _ = node.session.send.call_args

        assert req.body is None
        assert req.headers["accept-encoding"] == "gzip"
        print(req.headers)
        assert "content-encoding" not in req.headers

    @pytest.mark.parametrize("request_timeout", [None, 15])
    def test_timeout_override_default(self, request_timeout):
        node = self._get_mock_node(
            NodeConfig("http", "localhost", 80, request_timeout=request_timeout)
        )
        assert node.config.request_timeout == request_timeout

        node.perform_request("GET", "/")
        _, kwargs = node.session.send.call_args

        assert kwargs["timeout"] == request_timeout

        node.perform_request("GET", "/", request_timeout=5)
        _, kwargs = node.session.send.call_args

        assert kwargs["timeout"] == 5

        node.perform_request("GET", "/", request_timeout=None)
        _, kwargs = node.session.send.call_args

        assert kwargs["timeout"] is None

    def test_uses_https_if_verify_certs_is_off(self):
        with warnings.catch_warnings(record=True) as w:
            RequestsHttpNode(NodeConfig("https", "localhost", 443, verify_certs=False))

        assert 1 == len(w)
        assert (
            "Connecting to 'https://localhost:443' using TLS with verify_certs=False is insecure"
            == str(w[0].message)
        )

    def test_no_warn_when_uses_https_if_verify_certs_is_off(self):
        with warnings.catch_warnings(record=True) as w:
            RequestsHttpNode(
                NodeConfig(
                    "https", "localhost", 443, verify_certs=False, ssl_show_warn=False
                )
            )
        assert 0 == len(w)

    def test_no_warning_when_using_ssl_context(self):
        ctx = ssl.create_default_context()
        with warnings.catch_warnings(record=True) as w:
            RequestsHttpNode(NodeConfig("https", "localhost", 443, ssl_context=ctx))
        assert 0 == len(w)

    def test_ca_certs_with_verify_ssl_false_raises_error(self):
        with pytest.raises(ValueError) as e:
            RequestsHttpNode(
                NodeConfig(
                    "https", "localhost", 443, ca_certs="/ca/certs", verify_certs=False
                )
            )
        assert str(e.value) == "You cannot use 'ca_certs' when 'verify_certs=False'"

    def test_client_cert_is_used_as_session_cert(self):
        conn = RequestsHttpNode(
            NodeConfig("https", "localhost", 443, client_cert="/client/cert")
        )
        assert conn.session.cert == "/client/cert"

        conn = RequestsHttpNode(
            NodeConfig(
                "https",
                "localhost",
                443,
                client_cert="/client/cert",
                client_key="/client/key",
            )
        )
        assert conn.session.cert == ("/client/cert", "/client/key")

    def test_ca_certs_is_used_as_session_verify(self):
        conn = RequestsHttpNode(
            NodeConfig("https", "localhost", 443, ca_certs="/ca/certs")
        )
        assert conn.session.verify == "/ca/certs"

    def test_surrogatepass_into_bytes(self):
        data = b"\xe4\xbd\xa0\xe5\xa5\xbd\xed\xa9\xaa"
        node = self._get_mock_node(
            NodeConfig("http", "localhost", 80), response_body=data
        )
        _, data = node.perform_request("GET", "/")
        assert b"\xe4\xbd\xa0\xe5\xa5\xbd\xed\xa9\xaa" == data

    @pytest.mark.parametrize("_extras", [None, {}, {"requests.session.auth": None}])
    def test_requests_no_session_auth(self, _extras):
        node = self._get_mock_node(NodeConfig("http", "localhost", 80, _extras=_extras))
        assert node.session.auth is None

    def test_requests_custom_auth(self):
        auth = HTTPBasicAuth("username", "password")
        node = self._get_mock_node(
            NodeConfig("http", "localhost", 80, _extras={"requests.session.auth": auth})
        )
        assert node.session.auth is auth
        node.perform_request("GET", "/")
        (request,), _ = node.session.send.call_args
        assert request.headers["authorization"] == "Basic dXNlcm5hbWU6cGFzc3dvcmQ="

    def test_path_prefix(self):
        node = self._get_mock_node(
            NodeConfig(
                "http",
                "localhost",
                9200,
                path_prefix="/test",
            )
        )
        assert node.base_url == "http://localhost:9200/test"



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import gzip
import re
import ssl
import warnings
from unittest.mock import Mock, patch

import pytest
import urllib3
from urllib3.response import HTTPHeaderDict

from elastic_transport import NodeConfig, TransportError, Urllib3HttpNode
from elastic_transport._node._base import DEFAULT_USER_AGENT


class TestUrllib3HttpNode:
    def _get_mock_node(self, node_config, response_body=b"{}"):
        node = Urllib3HttpNode(node_config)

        def _dummy_urlopen(*args, **kwargs):
            dummy_response = Mock()
            dummy_response.headers = HTTPHeaderDict({})
            dummy_response.status = 200
            dummy_response.data = response_body
            _dummy_urlopen.call_args = (args, kwargs)
            return dummy_response

        node.pool.urlopen = _dummy_urlopen
        return node

    def test_close_pool(self):
        node = Urllib3HttpNode(NodeConfig("http", "localhost", 80))
        with patch.object(node.pool, "close") as pool_close:
            node.close()
        pool_close.assert_called_with()

    def test_ssl_context(self):
        ctx = ssl.create_default_context()
        node = Urllib3HttpNode(NodeConfig("https", "localhost", 80, ssl_context=ctx))
        assert len(node.pool.conn_kw.keys()) == 1
        assert isinstance(node.pool.conn_kw["ssl_context"], ssl.SSLContext)
        assert node.scheme == "https"

    def test_no_http_compression(self):
        node = self._get_mock_node(
            NodeConfig("http", "localhost", 80, http_compress=False)
        )
        assert not node.config.http_compress
        assert "accept-encoding" not in node.headers

        node.perform_request("GET", "/")
        (_, _), kwargs = node.pool.urlopen.call_args

        assert kwargs["body"] is None
        assert "accept-encoding" not in kwargs["headers"]
        assert "content-encoding" not in kwargs["headers"]

        node.perform_request("GET", "/", body=b"hello, world!")
        (_, _), kwargs = node.pool.urlopen.call_args

        assert kwargs["body"] == b"hello, world!"
        assert "accept-encoding" not in kwargs["headers"]
        assert "content-encoding" not in kwargs["headers"]

    @pytest.mark.parametrize(
        ["request_target", "expected_target"],
        [
            ("/_search", "/prefix/_search"),
            ("/?key=val", "/prefix/?key=val"),
            ("/_search?key=val/", "/prefix/_search?key=val/"),
        ],
    )
    def test_path_prefix_applied_to_target(self, request_target, expected_target):
        node = self._get_mock_node(
            NodeConfig("http", "localhost", 80, path_prefix="/prefix")
        )

        node.perform_request("GET", request_target)
        (_, target), _ = node.pool.urlopen.call_args

        assert target == expected_target

    @pytest.mark.parametrize("empty_body", [None, b""])
    def test_http_compression(self, empty_body):
        node = self._get_mock_node(
            NodeConfig("http", "localhost", 80, http_compress=True)
        )
        assert node.config.http_compress is True
        assert node.headers["accept-encoding"] == "gzip"

        # 'content-encoding' shouldn't be set at a connection level.
        # Should be applied only if the request is sent with a body.
        assert "content-encoding" not in node.headers

        node.perform_request("GET", "/", body=b"{}")

        (_, _), kwargs = node.pool.urlopen.call_args

        body = kwargs["body"]
        assert gzip.decompress(body) == b"{}"
        assert kwargs["headers"]["accept-encoding"] == "gzip"
        assert kwargs["headers"]["content-encoding"] == "gzip"

        node.perform_request("GET", "/", body=empty_body)

        (_, _), kwargs = node.pool.urlopen.call_args

        assert kwargs["body"] is None
        assert kwargs["headers"]["accept-encoding"] == "gzip"
        assert "content-encoding" not in kwargs["headers"]

    def test_default_headers(self):
        node = self._get_mock_node(NodeConfig("http", "localhost", 80))
        node.perform_request("GET", "/")
        (_, _), kwargs = node.pool.urlopen.call_args
        assert kwargs["headers"] == {
            "connection": "keep-alive",
            "user-agent": DEFAULT_USER_AGENT,
        }

    @pytest.mark.parametrize("request_timeout", [None, 15])
    def test_timeout_override_default(self, request_timeout):
        node = Urllib3HttpNode(
            NodeConfig("http", "localhost", 80, request_timeout=request_timeout)
        )
        assert node.config.request_timeout == request_timeout
        assert node.pool.timeout.total == request_timeout

        with patch.object(node.pool, "urlopen") as pool_urlopen:
            resp = Mock()
            resp.status = 200
            resp.headers = {}
            pool_urlopen.return_value = resp

            node.perform_request("GET", "/", request_timeout=request_timeout)
        _, kwargs = pool_urlopen.call_args

        assert kwargs["timeout"] == request_timeout

    def test_uses_https_if_verify_certs_is_off(self):
        with warnings.catch_warnings(record=True) as w:
            con = Urllib3HttpNode(
                NodeConfig("https", "localhost", 443, verify_certs=False)
            )
            assert 1 == len(w)
            assert (
                "Connecting to 'https://localhost:443' using TLS with verify_certs=False is insecure"
                == str(w[0].message)
            )

        assert isinstance(con.pool, urllib3.HTTPSConnectionPool)

    def test_no_warn_when_uses_https_if_verify_certs_is_off(self):
        with warnings.catch_warnings(record=True) as w:
            con = Urllib3HttpNode(
                NodeConfig(
                    "https", "localhost", 443, verify_certs=False, ssl_show_warn=False
                )
            )

        assert 0 == len(w)
        assert isinstance(con.pool, urllib3.HTTPSConnectionPool)

    def test_no_warning_when_using_ssl_context(self):
        ctx = ssl.create_default_context()
        with warnings.catch_warnings(record=True) as w:
            Urllib3HttpNode(NodeConfig("https", "localhost", 443, ssl_context=ctx))
        assert 0 == len(w)

    def test_surrogatepass_into_bytes(self):
        data = b"\xe4\xbd\xa0\xe5\xa5\xbd\xed\xa9\xaa"
        con = self._get_mock_node(
            NodeConfig("http", "localhost", 80), response_body=data
        )
        _, data = con.perform_request("GET", "/")
        assert b"\xe4\xbd\xa0\xe5\xa5\xbd\xed\xa9\xaa" == data

    @pytest.mark.xfail
    @patch("elastic_transport._node._base.logger")
    def test_uncompressed_body_logged(self, logger):
        con = self._get_mock_node(connection_params={"http_compress": True})
        con.perform_request("GET", "/", body=b'{"example": "body"}')

        assert 2 == logger.debug.call_count
        req, resp = logger.debug.call_args_list

        assert '> {"example": "body"}' == req[0][0] % req[0][1:]
        assert "< {}" == resp[0][0] % resp[0][1:]

    @pytest.mark.xfail
    @patch("elastic_transport._node._base.logger")
    def test_failed_request_logs(self, logger):
        conn = Urllib3HttpNode()

        with patch.object(conn.pool, "urlopen") as pool_urlopen:
            resp = Mock()
            resp.data = b'{"answer":42}'
            resp.status = 500
            resp.headers = {}
            pool_urlopen.return_value = resp

            with pytest.raises(TransportError) as e:
                conn.perform_request(
                    "GET",
                    "/?param=42",
                    b"{}",
                )

        assert repr(e.value) == "InternalServerError({'answer': 42}, status=500)"

        # log url and duration
        assert 1 == logger.warning.call_count
        assert re.match(
            r"^GET http://localhost/\?param=42 \[status:500 request:0.[0-9]{3}s\]",
            logger.warning.call_args[0][0] % logger.warning.call_args[0][1:],
        )
        assert 2 == logger.debug.call_count
        req, resp = logger.debug.call_args_list
        assert "> {}" == req[0][0] % req[0][1:]
        assert '< {"answer":42}' == resp[0][0] % resp[0][1:]



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import dataclasses
import json

import pytest

from elastic_transport import AsyncTransport, HttpxAsyncHttpNode
from elastic_transport._node._base import DEFAULT_USER_AGENT

from ..test_httpbin import parse_httpbin


@pytest.mark.anyio
async def test_simple_request(httpbin_node_config, httpbin):
    t = AsyncTransport([httpbin_node_config], node_class=HttpxAsyncHttpNode)

    resp, data = await t.perform_request(
        "GET",
        "/anything?key[]=1&key[]=2&q1&q2=",
        headers={"Custom": "headeR", "content-type": "application/json"},
        body={"JSON": "body"},
    )
    assert resp.status == 200
    assert data["method"] == "GET"
    assert data["url"] == f"{httpbin.url}/anything?key[]=1&key[]=2&q1&q2="

    # httpbin makes no-value query params into ''
    assert data["args"] == {
        "key[]": ["1", "2"],
        "q1": "",
        "q2": "",
    }
    assert data["data"] == '{"JSON":"body"}'
    assert data["json"] == {"JSON": "body"}

    request_headers = {
        "Content-Type": "application/json",
        "Content-Length": "15",
        "Custom": "headeR",
        "Connection": "keep-alive",
        "Host": f"{httpbin.host}:{httpbin.port}",
    }
    assert all(v == data["headers"][k] for k, v in request_headers.items())


@pytest.mark.anyio
async def test_node(httpbin_node_config, httpbin):
    def new_node(**kwargs):
        return HttpxAsyncHttpNode(dataclasses.replace(httpbin_node_config, **kwargs))

    node = new_node()
    resp, data = await node.perform_request("GET", "/anything")
    assert resp.status == 200
    parsed = parse_httpbin(data)
    assert parsed == {
        "headers": {
            "Accept": "*/*",
            "Accept-Encoding": "gzip, deflate, br",
            "Connection": "keep-alive",
            "Host": f"{httpbin.host}:{httpbin.port}",
            "User-Agent": DEFAULT_USER_AGENT,
        },
        "method": "GET",
        "url": f"{httpbin.url}/anything",
    }

    node = new_node(http_compress=True)
    resp, data = await node.perform_request("GET", "/anything")
    assert resp.status == 200
    parsed = parse_httpbin(data)
    assert parsed == {
        "headers": {
            "Accept": "*/*",
            "Accept-Encoding": "gzip",
            "Connection": "keep-alive",
            "Host": f"{httpbin.host}:{httpbin.port}",
            "User-Agent": DEFAULT_USER_AGENT,
        },
        "method": "GET",
        "url": f"{httpbin.url}/anything",
    }

    resp, data = await node.perform_request("GET", "/anything", body=b"hello, world!")
    assert resp.status == 200
    parsed = parse_httpbin(data)
    assert parsed == {
        "headers": {
            "Accept": "*/*",
            "Accept-Encoding": "gzip",
            "Content-Encoding": "gzip",
            "Content-Length": "33",
            "Connection": "keep-alive",
            "Host": f"{httpbin.host}:{httpbin.port}",
            "User-Agent": DEFAULT_USER_AGENT,
        },
        "method": "GET",
        "url": f"{httpbin.url}/anything",
    }

    resp, data = await node.perform_request(
        "POST",
        "/anything",
        body=json.dumps({"key": "value"}).encode("utf-8"),
        headers={"content-type": "application/json"},
    )
    assert resp.status == 200
    parsed = parse_httpbin(data)
    assert parsed == {
        "headers": {
            "Accept": "*/*",
            "Accept-Encoding": "gzip",
            "Content-Encoding": "gzip",
            "Content-Length": "36",
            "Content-Type": "application/json",
            "Connection": "keep-alive",
            "Host": f"{httpbin.host}:{httpbin.port}",
            "User-Agent": DEFAULT_USER_AGENT,
        },
        "method": "POST",
        "url": f"{httpbin.url}/anything",
    }



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import pytest

from elastic_transport import (
    AiohttpHttpNode,
    HttpxAsyncHttpNode,
    HttpxHttpNode,
    NodeConfig,
    RequestsHttpNode,
    Urllib3HttpNode,
)
from elastic_transport._node._base import ssl_context_from_node_config


@pytest.mark.parametrize(
    "node_cls",
    [
        Urllib3HttpNode,
        RequestsHttpNode,
        AiohttpHttpNode,
        HttpxHttpNode,
        HttpxAsyncHttpNode,
    ],
)
def test_unknown_parameter(node_cls):
    with pytest.raises(TypeError):
        node_cls(unknown_option=1)


@pytest.mark.parametrize(
    "host, check_hostname",
    [
        ("127.0.0.1", False),
        ("::1", False),
        ("localhost", True),
    ],
)
def test_ssl_context_from_node_config(host, check_hostname):
    node_config = NodeConfig("https", host, 443)
    ctx = ssl_context_from_node_config(node_config)
    assert ctx.check_hostname == check_hostname



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import pytest

from elastic_transport import ApiError, ApiResponseMeta, TransportError


def test_exception_repr_and_str():
    e = TransportError({"errors": [{"status": 500}]})
    assert repr(e) == "TransportError({'errors': [{'status': 500}]})"
    assert str(e) == "{'errors': [{'status': 500}]}"

    e = TransportError("error", errors=(ValueError("value error"),))
    assert repr(e) == "TransportError('error', errors={!r})".format(
        e.errors,
    )
    assert str(e) == "error"


def test_api_error_status_repr():
    e = ApiError(
        {"errors": [{"status": 500}]},
        body={},
        meta=ApiResponseMeta(
            status=500, http_version="1.1", headers={}, duration=0.0, node=None
        ),
    )
    assert (
        repr(e)
        == "ApiError({'errors': [{'status': 500}]}, meta=ApiResponseMeta(status=500, http_version='1.1', headers={}, duration=0.0, node=None), body={})"
    )
    assert str(e) == "[500] {'errors': [{'status': 500}]}"


def test_api_error_is_not_transport_error():
    with pytest.raises(ApiError):
        try:
            raise ApiError("", None, None)
        except TransportError:
            pass


def test_transport_error_is_not_api_error():
    with pytest.raises(TransportError):
        try:
            raise TransportError(
                "",
            )
        except ApiError:
            pass



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import hashlib
import logging
import socket
import ssl

import pytest
import trustme
from pytest_httpserver import HTTPServer

from elastic_transport import ApiResponseMeta, BaseNode, HttpHeaders, NodeConfig
from elastic_transport._node import NodeApiResponse


class DummyNode(BaseNode):
    def __init__(self, config: NodeConfig):
        super().__init__(config)
        self.exception = config._extras.pop("exception", None)
        self.status = config._extras.pop("status", 200)
        self.body = config._extras.pop("body", b"{}")
        self.calls = []
        self._headers = config._extras.pop("headers", {})

    def perform_request(self, *args, **kwargs):
        self.calls.append((args, kwargs))
        if self.exception:
            raise self.exception
        meta = ApiResponseMeta(
            node=self.config,
            duration=0.0,
            http_version="1.1",
            status=self.status,
            headers=HttpHeaders(self._headers),
        )
        return NodeApiResponse(meta, self.body)


class AsyncDummyNode(DummyNode):
    async def perform_request(self, *args, **kwargs):
        self.calls.append((args, kwargs))
        if self.exception:
            raise self.exception
        meta = ApiResponseMeta(
            node=self.config,
            duration=0.0,
            http_version="1.1",
            status=self.status,
            headers=HttpHeaders(self._headers),
        )
        return NodeApiResponse(meta, self.body)


@pytest.fixture(
    scope="session", params=["short-lower", "short-upper", "long-lower", "long-upper"]
)
def cert_fingerprint(request, httpbin_secure) -> str:
    """Gets the SHA256 fingerprint of the certificate for the secure httpbin"""
    sock = socket.create_connection((httpbin_secure.host, httpbin_secure.port))
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    sock = ctx.wrap_socket(sock)
    digest = hashlib.sha256(sock.getpeercert(binary_form=True)).hexdigest()
    assert len(digest) == 64
    sock.close()
    if "upper" in request.param:
        digest = digest.upper()
    else:
        digest = digest.lower()
    if "short" in request.param:
        return digest
    else:
        return ":".join([digest[i : i + 2] for i in range(0, len(digest), 2)])


@pytest.fixture(scope="session")
def httpbin_node_config(httpbin) -> NodeConfig:
    return NodeConfig(
        "http", httpbin.host, httpbin.port, verify_certs=False, ssl_show_warn=False
    )


@pytest.fixture(scope="function", autouse=True)
def elastic_transport_logging():
    for name in ("node", "node_pool", "transport"):
        logger = logging.getLogger(f"elastic_transport.{name}")
        for handler in logger.handlers[:]:
            logger.removeHandler(handler)


@pytest.fixture(scope="session")
def https_server_ip_node_config(tmp_path_factory: pytest.TempPathFactory) -> NodeConfig:
    ca = trustme.CA()
    tmpdir = tmp_path_factory.mktemp("certs")
    ca_cert_path = str(tmpdir / "ca.pem")
    ca.cert_pem.write_to_path(ca_cert_path)

    localhost_cert = ca.issue_cert("127.0.0.1")
    context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)

    crt = localhost_cert.cert_chain_pems[0]
    key = localhost_cert.private_key_pem
    with crt.tempfile() as crt_file, key.tempfile() as key_file:
        context.load_cert_chain(crt_file, key_file)

    server = HTTPServer(ssl_context=context)
    server.expect_request("/foobar").respond_with_json({"foo": "bar"})

    server.start()
    yield NodeConfig("https", "127.0.0.1", server.port, ca_certs=ca_cert_path)
    server.clear()
    if server.is_running():
        server.stop()



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from opentelemetry.sdk.trace import TracerProvider, export
from opentelemetry.sdk.trace.export.in_memory_span_exporter import InMemorySpanExporter

from elastic_transport import JsonSerializer
from elastic_transport._otel import OpenTelemetrySpan


def setup_tracing():
    tracer_provider = TracerProvider()
    memory_exporter = InMemorySpanExporter()
    span_processor = export.SimpleSpanProcessor(memory_exporter)
    tracer_provider.add_span_processor(span_processor)
    tracer = tracer_provider.get_tracer(__name__)

    return tracer, memory_exporter


def test_no_span():
    # With telemetry disabled, those calls should not raise
    span = OpenTelemetrySpan(None)
    span.set_db_statement(JsonSerializer().dumps({"timeout": "1m"}))
    span.set_node_metadata(
        "localhost",
        9200,
        "http://localhost:9200/",
        "_ml/anomaly_detectors/my-job/_open",
        "POST",
    )
    span.set_elastic_cloud_metadata(
        {
            "X-Found-Handling-Cluster": "e9106fc68e3044f0b1475b04bf4ffd5f",
            "X-Found-Handling-Instance": "instance-0000000001",
        }
    )


def test_detailed_span():
    tracer, memory_exporter = setup_tracing()
    with tracer.start_as_current_span("ml.open_job") as otel_span:
        span = OpenTelemetrySpan(
            otel_span,
            endpoint_id="my-job/_open",
            body_strategy="omit",
        )

        span.set_db_statement(JsonSerializer().dumps({"timeout": "1m"}))
        span.set_node_metadata(
            "localhost",
            9200,
            "http://localhost:9200/",
            "_ml/anomaly_detectors/my-job/_open",
            "POST",
        )
        span.set_elastic_cloud_metadata(
            {
                "X-Found-Handling-Cluster": "e9106fc68e3044f0b1475b04bf4ffd5f",
                "X-Found-Handling-Instance": "instance-0000000001",
            }
        )
        span.set_db_response(202)

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 1
    assert spans[0].name == "ml.open_job"
    assert spans[0].attributes == {
        "db.system.name": "elasticsearch",
        "url.full": "http://localhost:9200/_ml/anomaly_detectors/my-job/_open",
        "http.request.method": "POST",
        "server.address": "localhost",
        "server.port": 9200,
        "db.operation.name": "my-job/_open",
        "http.request.method": "POST",
        "db.namespace": "e9106fc68e3044f0b1475b04bf4ffd5f",
        "elasticsearch.node.name": "instance-0000000001",
        "db.response.status_code": "202",
    }


def test_db_statement():
    tracer, memory_exporter = setup_tracing()
    with tracer.start_as_current_span("search") as otel_span:
        span = OpenTelemetrySpan(otel_span, endpoint_id="search", body_strategy="raw")
        span.set_db_statement(JsonSerializer().dumps({"query": {"match_all": {}}}))

    spans = memory_exporter.get_finished_spans()
    assert len(spans) == 1
    assert spans[0].name == "search"
    assert spans[0].attributes == {
        "db.system.name": "elasticsearch",
        "db.operation.name": "search",
        "db.query.text": '{"query":{"match_all":{}}}',
    }



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import pytest

import elastic_transport
from elastic_transport import client_utils

modules = pytest.mark.parametrize("module", [elastic_transport, client_utils])


@modules
def test__all__sorted(module):
    module_all = module.__all__.copy()
    # Optional dependencies are added at the end
    if "OrjsonSerializer" in module_all:
        module_all.remove("OrjsonSerializer")
    assert module_all == sorted(module_all)


@modules
def test__all__is_importable(module):
    assert {attr for attr in module.__all__ if hasattr(module, attr)} == set(
        module.__all__
    )


def test_module_rewritten():
    assert repr(elastic_transport.Transport) == "<class 'elastic_transport.Transport'>"



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import pytest

from elastic_transport._utils import is_ipaddress


@pytest.mark.parametrize(
    "addr",
    [
        # IPv6
        "::1",
        "::",
        "FE80::8939:7684:D84b:a5A4%251",
        # IPv4
        "127.0.0.1",
        "8.8.8.8",
        b"127.0.0.1",
        # IPv6 w/ Zone IDs
        "FE80::8939:7684:D84b:a5A4%251",
        b"FE80::8939:7684:D84b:a5A4%251",
        "FE80::8939:7684:D84b:a5A4%19",
        b"FE80::8939:7684:D84b:a5A4%19",
    ],
)
def test_is_ipaddress(addr):
    assert is_ipaddress(addr)


@pytest.mark.parametrize(
    "addr",
    [
        "www.python.org",
        b"www.python.org",
        "v2.sg.media-imdb.com",
        b"v2.sg.media-imdb.com",
    ],
)
def test_is_not_ipaddress(addr):
    assert not is_ipaddress(addr)



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import pickle

import pytest

from elastic_transport import (
    ApiResponseMeta,
    BinaryApiResponse,
    HeadApiResponse,
    HttpHeaders,
    ListApiResponse,
    ObjectApiResponse,
    TextApiResponse,
)

meta = ApiResponseMeta(
    status=200, http_version="1.1", headers=HttpHeaders(), duration=0, node=None
)


@pytest.mark.parametrize(
    "response_cls",
    [TextApiResponse, BinaryApiResponse, ObjectApiResponse, ListApiResponse],
)
def test_response_meta(response_cls):
    resp = response_cls(meta=meta, body=None)
    assert resp.meta is meta

    assert resp == resp
    assert resp.body == resp
    assert resp == resp.body
    assert not resp != resp
    assert not (resp != resp.body)


def test_head_response():
    resp = HeadApiResponse(meta=meta)

    assert resp
    assert resp.body is True
    assert bool(resp) is True
    assert resp.meta is meta

    assert repr(resp) == "HeadApiResponse(True)"


def test_text_response():
    resp = TextApiResponse(body="Hello, world", meta=meta)
    assert resp.body == "Hello, world"
    assert len(resp) == 12
    assert resp.lower() == "hello, world"
    assert list(resp) == ["H", "e", "l", "l", "o", ",", " ", "w", "o", "r", "l", "d"]

    assert repr(resp) == "TextApiResponse('Hello, world')"


def test_binary_response():
    resp = BinaryApiResponse(body=b"Hello, world", meta=meta)
    assert resp.body == b"Hello, world"
    assert len(resp) == 12
    assert resp[0] == 72
    assert resp[:2] == b"He"
    assert resp.lower() == b"hello, world"
    assert resp.decode() == "Hello, world"
    assert list(resp) == [72, 101, 108, 108, 111, 44, 32, 119, 111, 114, 108, 100]

    assert repr(resp) == "BinaryApiResponse(b'Hello, world')"


def test_list_response():
    resp = ListApiResponse(body=[1, 2, 3], meta=meta)
    assert list(resp) == [1, 2, 3]
    assert resp.body == [1, 2, 3]
    assert resp[1] == 2

    assert repr(resp) == "ListApiResponse([1, 2, 3])"


def test_list_object_response():
    resp = ObjectApiResponse(body={"k1": 1, "k2": 2}, meta=meta)
    assert set(resp.keys()) == {"k1", "k2"}
    assert resp["k2"] == 2
    assert resp.body == {"k1": 1, "k2": 2}

    assert repr(resp) == "ObjectApiResponse({'k1': 1, 'k2': 2})"


@pytest.mark.parametrize(
    "resp_cls", [ObjectApiResponse, ListApiResponse, TextApiResponse, BinaryApiResponse]
)
@pytest.mark.parametrize(
    ["args", "kwargs"],
    [
        ((), {}),
        ((1,), {}),
        ((1,), {"raw": 1}),
        ((1,), {"body": 1}),
        ((1,), {"meta": 1}),
        ((), {"raw": 1, "body": 1}),
        ((), {"raw": 1, "body": 1, "meta": 1}),
        ((1,), {"raw": 1, "meta": 1}),
        ((1,), {"meta": 1, "body": 1}),
        ((1, 1), {"meta": 1, "body": 1}),
        ((), {"meta": 1, "body": 1, "unk": 1}),
    ],
)
def test_constructor_type_errors(resp_cls, args, kwargs):
    with pytest.raises(TypeError) as e:
        resp_cls(*args, **kwargs)
    assert str(e.value) == "Must pass 'meta' and 'body' to ApiResponse"


def test_constructor_allowed():
    resp = HeadApiResponse(meta)
    resp = HeadApiResponse(meta=meta)

    resp = ObjectApiResponse({}, meta)
    assert resp == {}

    resp = ObjectApiResponse(meta=meta, raw={})
    assert resp == {}

    resp = ObjectApiResponse(meta=meta, raw={}, body_cls=int)
    assert resp == {}

    resp = ObjectApiResponse(meta=meta, body={}, body_cls=int)
    assert resp == {}


@pytest.mark.parametrize(
    "response_cls, body",
    [
        (TextApiResponse, "Hello World"),
        (BinaryApiResponse, b"Hello World"),
        (ObjectApiResponse, {"Hello": "World"}),
        (ListApiResponse, ["Hello", "World"]),
    ],
)
def test_pickle(response_cls, body):
    resp = response_cls(meta=meta, body=body)
    pickled_resp = pickle.loads(pickle.dumps(resp))
    assert pickled_resp == resp
    assert pickled_resp.meta == resp.meta



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import random
import re
import ssl
import threading
import time
import warnings
from unittest import mock

import pytest

from elastic_transport import (
    AiohttpHttpNode,
    ConnectionError,
    ConnectionTimeout,
    HttpxHttpNode,
    NodeConfig,
    RequestsHttpNode,
    SniffingError,
    SniffOptions,
    Transport,
    TransportError,
    TransportWarning,
    Urllib3HttpNode,
)
from elastic_transport._transport import backoff_time
from elastic_transport.client_utils import DEFAULT
from tests.conftest import DummyNode


def test_transport_close_node_pool():
    t = Transport([NodeConfig("http", "localhost", 443)])
    with mock.patch.object(t.node_pool.all()[0], "close") as node_close:
        t.close()
    node_close.assert_called_with()


def test_request_with_custom_user_agent_header():
    t = Transport(
        [NodeConfig("http", "localhost", 80)], node_class=DummyNode, meta_header=False
    )

    t.perform_request("GET", "/", headers={"user-agent": "my-custom-value/1.2.3"})
    assert 1 == len(t.node_pool.get().calls)
    assert {
        "body": None,
        "request_timeout": DEFAULT,
        "headers": {"user-agent": "my-custom-value/1.2.3"},
    } == t.node_pool.get().calls[0][1]


def test_body_gets_encoded_into_bytes():
    t = Transport([NodeConfig("http", "localhost", 80)], node_class=DummyNode)

    t.perform_request(
        "GET", "/", headers={"Content-type": "application/json"}, body={"key": "你好"}
    )
    calls = t.node_pool.get().calls
    assert 1 == len(calls)
    args, kwargs = calls[0]
    assert ("GET", "/") == args
    assert kwargs["body"] == b'{"key":"\xe4\xbd\xa0\xe5\xa5\xbd"}'


def test_body_bytes_get_passed_untouched():
    t = Transport([NodeConfig("http", "localhost", 80)], node_class=DummyNode)

    body = b"\xe4\xbd\xa0\xe5\xa5\xbd"
    t.perform_request(
        "GET", "/", body=body, headers={"Content-Type": "application/json"}
    )
    calls = t.node_pool.get().calls
    assert 1 == len(calls)
    args, kwargs = calls[0]
    assert ("GET", "/") == args
    assert kwargs["body"] == b"\xe4\xbd\xa0\xe5\xa5\xbd"


def test_empty_response_with_content_type():
    t = Transport(
        [
            NodeConfig(
                "http",
                "localhost",
                80,
                _extras={"body": b"", "headers": {"Content-Type": "application/json"}},
            )
        ],
        node_class=DummyNode,
    )

    resp = t.perform_request("GET", "/", headers={"Accept": "application/json"})

    # Empty body is deserialized as 'None' instead of an error.
    assert resp.meta.status == 200
    assert resp.body is None


def test_kwargs_passed_on_to_node_pool():
    dt = object()
    t = Transport(
        [NodeConfig("http", "localhost", 80)],
        dead_node_backoff_factor=dt,
        max_dead_node_backoff=dt,
    )
    assert dt is t.node_pool.dead_node_backoff_factor
    assert dt is t.node_pool.max_dead_node_backoff


def test_request_will_fail_after_x_retries():
    t = Transport(
        [
            NodeConfig(
                "http",
                "localhost",
                80,
                _extras={"exception": ConnectionError("abandon ship")},
            )
        ],
        node_class=DummyNode,
        max_retries=0,
    )

    with pytest.raises(ConnectionError) as e:
        t.perform_request("GET", "/")

    assert 1 == len(t.node_pool.get().calls)
    assert len(e.value.errors) == 0

    # max_retries=3
    t = Transport(
        [
            NodeConfig(
                "http",
                "localhost",
                80,
                _extras={"exception": ConnectionError("abandon ship")},
            )
        ],
        node_class=DummyNode,
        max_retries=3,
    )

    with pytest.raises(ConnectionError) as e:
        t.perform_request("GET", "/")

    assert 4 == len(t.node_pool.get().calls)
    assert len(e.value.errors) == 3
    assert all(isinstance(error, ConnectionError) for error in e.value.errors)

    # max_retries=2 in perform_request()
    with pytest.raises(ConnectionError) as e:
        t.perform_request("GET", "/", max_retries=2)

    assert 7 == len(t.node_pool.get().calls)
    assert len(e.value.errors) == 2
    assert all(isinstance(error, ConnectionError) for error in e.value.errors)


@pytest.mark.parametrize("retry_on_timeout", [True, False])
def test_retry_on_timeout(retry_on_timeout):
    t = Transport(
        [
            NodeConfig(
                "http",
                "localhost",
                80,
                _extras={"exception": ConnectionTimeout("abandon ship")},
            ),
            NodeConfig(
                "http",
                "localhost",
                81,
                _extras={"exception": ConnectionError("error!")},
            ),
        ],
        node_class=DummyNode,
        max_retries=1,
        retry_on_timeout=retry_on_timeout,
        randomize_nodes_in_pool=False,
    )

    if retry_on_timeout:
        with pytest.raises(ConnectionError) as e:
            t.perform_request("GET", "/")
        assert len(e.value.errors) == 1
        assert isinstance(e.value.errors[0], ConnectionTimeout)

    else:
        with pytest.raises(ConnectionTimeout) as e:
            t.perform_request("GET", "/")
        assert len(e.value.errors) == 0


def test_retry_on_status():
    t = Transport(
        [
            NodeConfig("http", "localhost", 80, _extras={"status": 404}),
            NodeConfig(
                "http",
                "localhost",
                81,
                _extras={"status": 401},
            ),
            NodeConfig(
                "http",
                "localhost",
                82,
                _extras={"status": 403},
            ),
            NodeConfig(
                "http",
                "localhost",
                83,
                _extras={"status": 555},
            ),
        ],
        node_class=DummyNode,
        node_selector_class="round_robin",
        retry_on_status=(401, 403, 404),
        randomize_nodes_in_pool=False,
        max_retries=5,
    )

    meta, _ = t.perform_request("GET", "/")
    assert meta.status == 555

    # Assert that every node is called once
    node_calls = [len(node.calls) for node in t.node_pool.all()]
    assert node_calls == [
        1,
        1,
        1,
        1,
    ]


def test_request_retry_backoff():
    t = Transport(
        [
            NodeConfig(
                "http",
                "localhost",
                80,
                _extras={"exception": ConnectionError("abandon ship")},
            )
        ],
        node_class=DummyNode,
        max_retries=3,
        retry_backoff_base=1,
        retry_backoff_cap=2,
    )

    with mock.patch("elastic_transport._transport.time.sleep") as mock_sleep:
        with pytest.raises(ConnectionError) as e:
            t.perform_request("GET", "/")

    assert 4 == len(t.node_pool.get().calls)
    assert len(e.value.errors) == 3
    assert all(isinstance(error, ConnectionError) for error in e.value.errors)

    assert mock_sleep.call_count == 3
    assert all(0 < arg[0][0] <= 2 for arg in mock_sleep.call_args_list)


def test_failed_connection_will_be_marked_as_dead():
    t = Transport(
        [
            NodeConfig(
                "http",
                "localhost",
                80,
                _extras={"exception": ConnectionError("abandon ship")},
            ),
            NodeConfig(
                "http",
                "localhost",
                81,
                _extras={"exception": ConnectionError("abandon ship")},
            ),
        ],
        max_retries=3,
        node_class=DummyNode,
    )

    with pytest.raises(ConnectionError) as e:
        t.perform_request("GET", "/")
    assert 0 == len(t.node_pool._alive_nodes)
    assert 2 == len(t.node_pool._dead_nodes.queue)
    assert len(e.value.errors) == 3
    assert all(isinstance(error, ConnectionError) for error in e.value.errors)


def test_resurrected_connection_will_be_marked_as_live_on_success():
    for method in ("GET", "HEAD"):
        t = Transport(
            [
                NodeConfig("http", "localhost", 80),
                NodeConfig("http", "localhost", 81),
            ],
            node_class=DummyNode,
        )
        node1 = t.node_pool.get()
        node2 = t.node_pool.get()
        t.node_pool.mark_dead(node1)
        t.node_pool.mark_dead(node2)

        t.perform_request(method, "/")
        assert 1 == len(t.node_pool._alive_nodes)
        assert 1 == len(t.node_pool._dead_consecutive_failures)
        assert 1 == len(t.node_pool._dead_nodes.queue)


def test_sniff_on_node_failure_error_doesnt_raise():
    t = Transport(
        [
            NodeConfig("http", "localhost", 80, _extras={"status": 502}),
            NodeConfig("http", "localhost", 81),
        ],
        max_retries=1,
        retry_on_status=(502,),
        node_class=DummyNode,
        randomize_nodes_in_pool=False,
    )
    bad_node = t.node_pool._all_nodes[NodeConfig("http", "localhost", 80)]
    with (
        mock.patch.object(t, "sniff") as sniff,
        mock.patch.object(t.node_pool, "mark_dead") as mark_dead,
    ):
        sniff.side_effect = TransportError("sniffing error!")
        t.perform_request("GET", "/")
    mark_dead.assert_called_with(bad_node)


def test_node_class_as_string():
    t = Transport([NodeConfig("http", "localhost", 80)], node_class="urllib3")
    assert isinstance(t.node_pool.get(), Urllib3HttpNode)

    t = Transport([NodeConfig("http", "localhost", 80)], node_class="requests")
    assert isinstance(t.node_pool.get(), RequestsHttpNode)

    with pytest.raises(ValueError) as e:
        Transport([NodeConfig("http", "localhost", 80)], node_class="huh?")
    assert str(e.value) == (
        "Unknown option for node_class: 'huh?'. "
        "Available options are: 'aiohttp', 'httpx', 'httpxasync', 'requests', 'urllib3'"
    )


@pytest.mark.parametrize(["status", "boolean"], [(200, True), (299, True)])
def test_head_response_true(status, boolean):
    t = Transport(
        [NodeConfig("http", "localhost", 80, _extras={"status": status, "body": b""})],
        node_class=DummyNode,
    )
    resp, data = t.perform_request("HEAD", "/")
    assert resp.status == status
    assert data is None


def test_head_response_false():
    t = Transport(
        [NodeConfig("http", "localhost", 80, _extras={"status": 404, "body": b""})],
        node_class=DummyNode,
    )
    meta, resp = t.perform_request("HEAD", "/")
    assert meta.status == 404
    assert resp is None
    # 404s don't count as a dead node status.
    assert 0 == len(t.node_pool._dead_nodes.queue)


@pytest.mark.parametrize(
    "node_class",
    ["urllib3", "requests", "httpx", Urllib3HttpNode, RequestsHttpNode, HttpxHttpNode],
)
def test_transport_client_meta_node_class(node_class):
    t = Transport([NodeConfig("http", "localhost", 80)], node_class=node_class)
    assert (
        t._transport_client_meta[3] == t.node_pool.node_class._CLIENT_META_HTTP_CLIENT
    )
    assert t._transport_client_meta[3][0] in ("ur", "rq", "hx")
    assert re.match(
        r"^et=[0-9.]+p?,py=[0-9.]+p?,t=[0-9.]+p?,(?:ur|rq|hx)=[0-9.]+p?$",
        ",".join(f"{k}={v}" for k, v in t._transport_client_meta),
    )

    # Defaults to urllib3
    t = Transport([NodeConfig("http", "localhost", 80)])
    assert t._transport_client_meta[3][0] == "ur"
    assert [x[0] for x in t._transport_client_meta[:3]] == ["et", "py", "t"]


@pytest.mark.parametrize(
    "node_class",
    ["aiohttp", AiohttpHttpNode],
)
def test_transport_and_node_are_sync(node_class):
    with pytest.raises(ValueError) as e:
        Transport([NodeConfig("http", "localhost", 80)], node_class=node_class)
    assert str(e.value) == "Specified 'node_class' is async, should be sync instead"


def test_client_meta_header():
    class DummyNodeWithClientMeta(DummyNode):
        _CLIENT_META_HTTP_CLIENT = ("dm", "0.0.0p")

    t = Transport(
        [NodeConfig("http", "localhost", 80)],
        node_class=DummyNodeWithClientMeta,
        client_meta_service=("es", "8.0.0p"),
    )
    t.perform_request("GET", "/")

    calls = t.node_pool.get().calls
    assert 1 == len(calls)
    headers = calls[0][1]["headers"]

    assert sorted(headers.keys()) == ["x-elastic-client-meta"]
    assert re.match(
        r"^es=8\.0\.0p,py=[0-9.]+p?,t=[0-9.]+p?,dm=0\.0\.0p$",
        headers["x-elastic-client-meta"],
    )


def test_client_meta_header_extras():
    class DummyNodeWithClientMeta(DummyNode):
        _CLIENT_META_HTTP_CLIENT = ("dm", "0.0.0p")

    t = Transport(
        [NodeConfig("http", "localhost", 80)],
        node_class=DummyNodeWithClientMeta,
        client_meta_service=("es", "8.0.0p"),
    )
    t.perform_request("GET", "/", client_meta=(("h", "s"),))

    calls = t.node_pool.get().calls
    assert 1 == len(calls)
    headers = calls[0][1]["headers"]

    assert sorted(headers.keys()) == ["x-elastic-client-meta"]
    assert re.match(
        r"^es=8\.0\.0p,py=[0-9.]+p?,t=[0-9.]+p?,dm=0\.0\.0p,h=s$",
        headers["x-elastic-client-meta"],
    )


def test_sniff_on_start():
    calls = []

    def sniff_callback(*args):
        calls.append(args)
        return [NodeConfig("http", "localhost", 80)]

    t = Transport(
        [NodeConfig("http", "localhost", 80)],
        node_class=DummyNode,
        sniff_on_start=True,
        sniff_callback=sniff_callback,
    )
    assert len(calls) == 1

    t.perform_request("GET", "/")

    assert len(calls) == 1
    transport, sniff_options = calls[0]
    assert transport is t
    assert sniff_options == SniffOptions(is_initial_sniff=True, sniff_timeout=0.5)


def test_sniff_before_requests():
    calls = []

    def sniff_callback(*args):
        calls.append(args)
        return []

    t = Transport(
        [NodeConfig("http", "localhost", 80)],
        node_class=DummyNode,
        sniff_before_requests=True,
        sniff_callback=sniff_callback,
    )
    assert len(calls) == 0

    t.perform_request("GET", "/")

    assert len(calls) == 1
    transport, sniff_options = calls[0]
    assert transport is t
    assert sniff_options == SniffOptions(is_initial_sniff=False, sniff_timeout=0.5)


def test_sniff_on_node_failure():
    calls = []

    def sniff_callback(*args):
        calls.append(args)
        return []

    t = Transport(
        [
            NodeConfig("http", "localhost", 80),
            NodeConfig("http", "localhost", 81, _extras={"status": 500}),
        ],
        randomize_nodes_in_pool=False,
        node_selector_class="round_robin",
        node_class=DummyNode,
        max_retries=1,
        sniff_on_node_failure=True,
        sniff_callback=sniff_callback,
    )
    assert len(calls) == 0

    t.perform_request("GET", "/")  # 200
    assert len(calls) == 0

    t.perform_request("GET", "/")  # 500
    assert len(calls) == 1

    transport, sniff_options = calls[0]
    assert transport is t
    assert sniff_options == SniffOptions(is_initial_sniff=False, sniff_timeout=0.5)


@pytest.mark.parametrize(
    "kwargs",
    [
        {"sniff_on_start": True},
        {"sniff_on_node_failure": True},
        {"sniff_before_requests": True},
    ],
)
def test_error_with_sniffing_enabled_without_callback(kwargs):
    with pytest.raises(ValueError) as e:
        Transport([NodeConfig("http", "localhost", 80)], **kwargs)

    assert str(e.value) == "Enabling sniffing requires specifying a 'sniff_callback'"


def test_error_sniffing_callback_without_sniffing_enabled():
    with pytest.raises(ValueError) as e:
        Transport([NodeConfig("http", "localhost", 80)], sniff_callback=lambda *_: [])

    assert str(e.value) == (
        "Using 'sniff_callback' requires enabling sniffing via 'sniff_on_start', "
        "'sniff_before_requests' or 'sniff_on_node_failure'"
    )


def test_heterogeneous_node_config_warning_with_sniffing():
    with warnings.catch_warnings(record=True) as w:
        context = ssl.create_default_context()
        Transport(
            [
                NodeConfig(
                    "https", "localhost", 80, path_prefix="/a", ssl_context=context
                ),
                NodeConfig(
                    "https", "localhost", 81, path_prefix="/b", ssl_context=context
                ),
            ],
            sniff_on_start=True,
            sniff_callback=lambda *_: [
                NodeConfig("https", "localhost", 80, path_prefix="/a")
            ],
        )

    assert len(w) == 1
    assert w[0].category == TransportWarning
    assert str(w[0].message) == (
        "Detected NodeConfig instances with different options. It's "
        "recommended to keep all options except for 'host' and 'port' "
        "the same for sniffing to work reliably."
    )


def test_sniffed_nodes_added_to_pool():
    sniffed_nodes = [
        NodeConfig("http", "localhost", 80),
        NodeConfig("http", "localhost", 81),
    ]

    t = Transport(
        [
            NodeConfig("http", "localhost", 80),
        ],
        node_class=DummyNode,
        sniff_before_requests=True,
        sniff_callback=lambda *_: sniffed_nodes,
    )
    assert len(t.node_pool) == 1

    t.perform_request("GET", "/")

    # The node pool knows when nodes are already in the pool
    # so we shouldn't get duplicates after sniffing.
    assert len(t.node_pool) == 2
    assert set(sniffed_nodes) == {node.config for node in t.node_pool.all()}


def test_sniff_error_resets_lock_and_last_sniffed_at():
    def sniff_error(*_):
        raise TransportError("This is an error!")

    t = Transport(
        [
            NodeConfig("http", "localhost", 80),
        ],
        node_class=DummyNode,
        sniff_before_requests=True,
        sniff_callback=sniff_error,
    )
    last_sniffed_at = t._last_sniffed_at

    with pytest.raises(TransportError) as e:
        t.perform_request("GET", "/")
    assert str(e.value) == "This is an error!"

    assert t._last_sniffed_at == last_sniffed_at
    assert t._sniffing_lock.locked() is False


def test_sniff_on_start_no_results_errors():
    with pytest.raises(SniffingError) as e:
        Transport(
            [
                NodeConfig("http", "localhost", 80),
            ],
            node_class=DummyNode,
            sniff_on_start=True,
            sniff_callback=lambda *_: [],
        )

    assert (
        str(e.value) == "No viable nodes were discovered on the initial sniff attempt"
    )


@pytest.mark.parametrize("pool_size", [1, 8])
def test_threading_test(pool_size):
    node_configs = [
        NodeConfig("http", "localhost", 80),
        NodeConfig("http", "localhost", 81),
        NodeConfig("http", "localhost", 82),
        NodeConfig("http", "localhost", 83, _extras={"status": 500}),
    ]

    def sniff_callback(*_):
        time.sleep(random.random())
        return node_configs

    t = Transport(
        node_configs,
        retry_on_status=[500],
        max_retries=5,
        node_class=DummyNode,
        sniff_on_start=True,
        sniff_before_requests=True,
        sniff_on_node_failure=True,
        sniff_callback=sniff_callback,
    )

    class ThreadTest(threading.Thread):
        def __init__(self):
            super().__init__()
            self.successful_requests = 0

        def run(self) -> None:
            while time.time() < start + 2:
                t.perform_request("GET", "/")
                self.successful_requests += 1

    threads = [ThreadTest() for _ in range(pool_size * 2)]
    start = time.time()
    [thread.start() for thread in threads]
    [thread.join() for thread in threads]

    assert sum(thread.successful_requests for thread in threads) >= 1000


def test_httpbin(httpbin_node_config):
    t = Transport([httpbin_node_config])
    resp = t.perform_request("GET", "/anything")
    assert resp.meta.status == 200
    assert isinstance(resp.body, dict)


def test_backoff_time():
    for i in range(1, 11):
        assert backoff_time(i, 0, 0) == 0
    for i in range(1, 11):
        assert backoff_time(i, 0, 1) == 0
    exp = [1, 2, 4, 8, 16, 32, 64, 128, 256, 512]
    for i in range(10):
        ceiling = min(1, 0.1 * exp[i])
        assert ceiling / 2 <= backoff_time(i + 1, 0.1, 1) <= ceiling
    for i in range(10):
        ceiling = min(1, exp[i])
        assert ceiling / 2 <= backoff_time(i + 1, 1, 1) <= ceiling
    for i in range(10):
        ceiling = min(100, exp[i])
        assert ceiling / 2 <= backoff_time(i + 1, 1, 100) <= ceiling
    for i in range(10):
        ceiling = min(600, 60 * exp[i])
        assert ceiling / 2 <= backoff_time(i + 1, 60, 600) <= ceiling



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import sys
import warnings

import pytest

from elastic_transport import NodeConfig, RequestsHttpNode, TlsError, Urllib3HttpNode

requires_ssl_assert_fingerprint_in_chain = pytest.mark.skipif(
    sys.version_info < (3, 10) or sys.implementation.name != "cpython",
    reason="Requires CPython 3.10+",
)


@requires_ssl_assert_fingerprint_in_chain
@pytest.mark.parametrize("node_cls", [Urllib3HttpNode, RequestsHttpNode])
def test_ssl_assert_fingerprint_invalid_length(node_cls, httpbin_secure):
    with pytest.raises(ValueError) as e:
        node_cls(
            NodeConfig(
                "https",
                httpbin_secure.host,
                httpbin_secure.port,
                ssl_assert_fingerprint="0000",
            )
        )

    assert (
        str(e.value)
        == "Fingerprint of invalid length '4', should be one of '32', '40', '64'"
    )


@requires_ssl_assert_fingerprint_in_chain
@pytest.mark.parametrize("node_cls", [Urllib3HttpNode, RequestsHttpNode])
def test_assert_fingerprint_in_cert_chain(node_cls, cert_fingerprint, httpbin_secure):
    with warnings.catch_warnings(record=True) as w:
        node = node_cls(
            NodeConfig(
                "https",
                httpbin_secure.host,
                httpbin_secure.port,
                ssl_assert_fingerprint=cert_fingerprint,
            )
        )
        meta, _ = node.perform_request("GET", "/")
        assert meta.status == 200

    assert w == []


@requires_ssl_assert_fingerprint_in_chain
@pytest.mark.parametrize("node_cls", [Urllib3HttpNode, RequestsHttpNode])
def test_assert_fingerprint_in_cert_chain_failure(
    node_cls, httpbin_secure, cert_fingerprint
):
    node = node_cls(
        NodeConfig(
            "https",
            "www.elastic.co",
            443,
            ssl_assert_fingerprint="0" * 64,
        )
    )

    with pytest.raises(TlsError) as e:
        node.perform_request("GET", "/")

    err = str(e.value)
    assert "Fingerprints did not match." in err
    # This is the bad value we "expected"
    assert (
        'Expected "0000000000000000000000000000000000000000000000000000000000000000",'
        in err
    )
    # This is the root CA for www.elastic.co with a leading comma to denote more than one cert was listed.
    assert ', "cbb522d7b7f127ad6a0113865bdf1cd4102e7d0759af635a7cf4720dc963c53b"' in err



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import warnings

import pytest

from elastic_transport import AsyncTransport


@pytest.mark.asyncio
async def test_simple_request(https_server_ip_node_config):
    with warnings.catch_warnings():
        warnings.simplefilter("error")

        t = AsyncTransport([https_server_ip_node_config])

        resp, data = await t.perform_request("GET", "/foobar")
        assert resp.status == 200
        assert data == {"foo": "bar"}



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import functools
import socket
import ssl

import pytest

from elastic_transport import (
    AiohttpHttpNode,
    ConnectionError,
    HttpxAsyncHttpNode,
    HttpxHttpNode,
    NodeConfig,
    RequestsHttpNode,
    TlsError,
    Urllib3HttpNode,
)
from elastic_transport._compat import await_if_coro
from elastic_transport.client_utils import url_to_node_config

TLSv1_0_URL = "https://tls-v1-0.badssl.com:1010"
TLSv1_1_URL = "https://tls-v1-1.badssl.com:1011"
TLSv1_2_URL = "https://tls-v1-2.badssl.com:1012"

node_classes = pytest.mark.parametrize(
    "node_class",
    [
        AiohttpHttpNode,
        Urllib3HttpNode,
        RequestsHttpNode,
        HttpxHttpNode,
        HttpxAsyncHttpNode,
    ],
)

supported_version_params = [
    (TLSv1_0_URL, ssl.PROTOCOL_TLSv1),
    (TLSv1_1_URL, ssl.PROTOCOL_TLSv1_1),
    (TLSv1_2_URL, ssl.PROTOCOL_TLSv1_2),
    (TLSv1_2_URL, None),
]
unsupported_version_params = [
    (TLSv1_0_URL, None),
    (TLSv1_1_URL, None),
    (TLSv1_0_URL, ssl.PROTOCOL_TLSv1_1),
    (TLSv1_0_URL, ssl.PROTOCOL_TLSv1_2),
    (TLSv1_1_URL, ssl.PROTOCOL_TLSv1_2),
]

try:
    from ssl import TLSVersion
except ImportError:
    pass
else:
    supported_version_params.extend(
        [
            (TLSv1_0_URL, TLSVersion.TLSv1),
            (TLSv1_1_URL, TLSVersion.TLSv1_1),
            (TLSv1_2_URL, TLSVersion.TLSv1_2),
        ]
    )
    unsupported_version_params.extend(
        [
            (TLSv1_0_URL, TLSVersion.TLSv1_1),
            (TLSv1_0_URL, TLSVersion.TLSv1_2),
            (TLSv1_1_URL, TLSVersion.TLSv1_2),
            (TLSv1_0_URL, TLSVersion.TLSv1_3),
            (TLSv1_1_URL, TLSVersion.TLSv1_3),
            (TLSv1_2_URL, TLSVersion.TLSv1_3),
        ]
    )


@functools.lru_cache()
def tlsv1_1_supported() -> bool:
    # OpenSSL distributions on Ubuntu/Debian disable TLSv1.1 and before incorrectly.
    # So we try to detect that and skip tests when needed.
    try:
        sock = socket.create_connection(("tls-v1-1.badssl.com", 1011))
        ctx = ssl.SSLContext(ssl.PROTOCOL_TLSv1_1)
        sock = ctx.wrap_socket(sock, server_hostname="tls-v1-1.badssl.com")
        sock.close()
    except ssl.SSLError:
        return False
    return True


@node_classes
@pytest.mark.parametrize(
    ["url", "ssl_version"],
    supported_version_params,
)
@pytest.mark.anyio
async def test_supported_tls_versions(
    node_class, url: str, ssl_version: int, anyio_backend
):
    if url in (TLSv1_0_URL, TLSv1_1_URL) and not tlsv1_1_supported():
        pytest.skip("TLSv1.1 isn't supported by this OpenSSL distribution")
    if anyio_backend == "trio" and node_class is not HttpxAsyncHttpNode:
        pytest.skip("only httpx supports trio")

    node_config = url_to_node_config(url).replace(ssl_version=ssl_version)
    node = node_class(node_config)

    resp, _ = await await_if_coro(node.perform_request("GET", "/"))
    assert resp.status == 200


@node_classes
@pytest.mark.parametrize(
    ["url", "ssl_version"],
    unsupported_version_params,
)
@pytest.mark.anyio
async def test_unsupported_tls_versions(
    node_class, url: str, ssl_version: int, anyio_backend
):
    if anyio_backend == "trio" and node_class is not HttpxAsyncHttpNode:
        pytest.skip("only httpx supports trio")

    node_config = url_to_node_config(url).replace(ssl_version=ssl_version)
    node = node_class(node_config)

    # Remove ConnectionError when we have a fix or workaround for
    # https://github.com/encode/httpx/discussions/3674
    with pytest.raises((TlsError, ConnectionError)) as e:
        await await_if_coro(node.perform_request("GET", "/"))
    if anyio_backend == "trio" and node_class is HttpxAsyncHttpNode:
        return  # Trio errors are not correctly bubbled up by httpx
    assert "unsupported protocol" in str(e.value) or "handshake failure" in str(e.value)


@node_classes
@pytest.mark.parametrize("ssl_version", [0, "TLSv1", object()])
def test_ssl_version_value_error(node_class, ssl_version):
    # Remove ConnectionError when we have a fix or workaround for
    # https://github.com/encode/httpx/discussions/3674
    with pytest.raises((ValueError, ConnectionError)) as e:
        node_class(NodeConfig("https", "localhost", 9200, ssl_version=ssl_version))
    assert str(e.value) == (
        f"Unsupported value for 'ssl_version': {ssl_version!r}. Must be either "
        "'ssl.PROTOCOL_TLSvX' or 'ssl.TLSVersion.TLSvX'"
    )



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import uuid
from datetime import date
from decimal import Decimal

import pytest

from elastic_transport import (
    JsonSerializer,
    NdjsonSerializer,
    OrjsonSerializer,
    SerializationError,
    SerializerCollection,
    TextSerializer,
)
from elastic_transport._serializer import DEFAULT_SERIALIZERS

serializers = SerializerCollection(DEFAULT_SERIALIZERS)


@pytest.fixture(params=[JsonSerializer, OrjsonSerializer])
def json_serializer(request: pytest.FixtureRequest):
    yield request.param()


def test_date_serialization(json_serializer):
    assert b'{"d":"2010-10-01"}' == json_serializer.dumps({"d": date(2010, 10, 1)})


def test_decimal_serialization(json_serializer):
    assert b'{"d":3.8}' == json_serializer.dumps({"d": Decimal("3.8")})


def test_uuid_serialization(json_serializer):
    assert b'{"d":"00000000-0000-0000-0000-000000000003"}' == json_serializer.dumps(
        {"d": uuid.UUID("00000000-0000-0000-0000-000000000003")}
    )


def test_serializes_nan():
    assert b'{"d":NaN}' == JsonSerializer().dumps({"d": float("NaN")})
    # NaN is invalid JSON, and orjson silently converts it to null
    assert b'{"d":null}' == OrjsonSerializer().dumps({"d": float("NaN")})


def test_raises_serialization_error_on_dump_error(json_serializer):
    with pytest.raises(SerializationError):
        json_serializer.dumps(object())
    with pytest.raises(SerializationError):
        TextSerializer().dumps({})


def test_raises_serialization_error_on_load_error(json_serializer):
    with pytest.raises(SerializationError):
        json_serializer.loads(object())
    with pytest.raises(SerializationError):
        json_serializer.loads(b"{{")


def test_json_unicode_is_handled(json_serializer):
    assert (
        json_serializer.dumps({"你好": "你好"})
        == b'{"\xe4\xbd\xa0\xe5\xa5\xbd":"\xe4\xbd\xa0\xe5\xa5\xbd"}'
    )
    assert json_serializer.loads(
        b'{"\xe4\xbd\xa0\xe5\xa5\xbd":"\xe4\xbd\xa0\xe5\xa5\xbd"}'
    ) == {"你好": "你好"}


def test_text_unicode_is_handled():
    text_serializer = TextSerializer()
    assert text_serializer.dumps("你好") == b"\xe4\xbd\xa0\xe5\xa5\xbd"
    assert text_serializer.loads(b"\xe4\xbd\xa0\xe5\xa5\xbd") == "你好"


def test_json_unicode_surrogates_handled():
    assert (
        JsonSerializer().dumps({"key": "你好\uda6a"})
        == b'{"key":"\xe4\xbd\xa0\xe5\xa5\xbd\xed\xa9\xaa"}'
    )
    assert JsonSerializer().loads(
        b'{"key":"\xe4\xbd\xa0\xe5\xa5\xbd\xed\xa9\xaa"}'
    ) == {"key": "你好\uda6a"}

    # orjson is strict about UTF-8
    with pytest.raises(SerializationError):
        OrjsonSerializer().dumps({"key": "你好\uda6a"})

    with pytest.raises(SerializationError):
        OrjsonSerializer().loads(b'{"key":"\xe4\xbd\xa0\xe5\xa5\xbd\xed\xa9\xaa"}')


def test_text_unicode_surrogates_handled(json_serializer):
    text_serializer = TextSerializer()
    assert (
        text_serializer.dumps("你好\uda6a") == b"\xe4\xbd\xa0\xe5\xa5\xbd\xed\xa9\xaa"
    )
    assert (
        text_serializer.loads(b"\xe4\xbd\xa0\xe5\xa5\xbd\xed\xa9\xaa") == "你好\uda6a"
    )


def test_deserializes_json_by_default():
    assert {"some": "data"} == serializers.loads(b'{"some":"data"}')


def test_deserializes_text_with_correct_ct():
    assert '{"some":"data"}' == serializers.loads(b'{"some":"data"}', "text/plain")
    assert '{"some":"data"}' == serializers.loads(
        b'{"some":"data"}', "text/plain; charset=whatever"
    )


def test_raises_serialization_error_on_unknown_mimetype():
    with pytest.raises(SerializationError) as e:
        serializers.loads(b"{}", "fake/type")
    assert (
        str(e.value)
        == "Unknown mimetype, not able to serialize or deserialize: fake/type"
    )


def test_raises_improperly_configured_when_default_mimetype_cannot_be_deserialized():
    with pytest.raises(ValueError) as e:
        SerializerCollection({})
    assert (
        str(e.value)
        == "Must configure a serializer for the default mimetype 'application/json'"
    )


def test_text_asterisk_works_for_all_text_types():
    assert serializers.loads(b"{}", "text/html") == "{}"
    assert serializers.dumps("{}", "text/html") == b"{}"


@pytest.mark.parametrize("should_strip", [False, b"\n", b"\r\n"])
def test_ndjson_loads(should_strip):
    serializer = NdjsonSerializer()
    data = (
        b'{"key":"value"}\n'
        b'{"number":0.1,"one":1}\n'
        b'{"list":[1,2,3]}\r\n'
        b'{"unicode":"\xe4\xbd\xa0\xe5\xa5\xbd\xed\xa9\xaa"}\r\n'
    )
    if should_strip:
        data = data.strip(should_strip)
    data = serializer.loads(data)

    assert data == [
        {"key": "value"},
        {"number": 0.1, "one": 1},
        {"list": [1, 2, 3]},
        {"unicode": "你好\uda6a"},
    ]


def test_ndjson_dumps():
    serializer = NdjsonSerializer()
    data = serializer.dumps(
        [
            {"key": "value"},
            {"number": 0.1, "one": 1},
            {"list": [1, 2, 3]},
            {"unicode": "你好\uda6a"},
            '{"key:"value"}',
            b'{"bytes":"too"}',
        ]
    )
    assert data == (
        b'{"key":"value"}\n'
        b'{"number":0.1,"one":1}\n'
        b'{"list":[1,2,3]}\n'
        b'{"unicode":"\xe4\xbd\xa0\xe5\xa5\xbd\xed\xa9\xaa"}\n'
        b'{"key:"value"}\n'
        b'{"bytes":"too"}\n'
    )



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import io
import logging

import pytest

from elastic_transport import (
    AiohttpHttpNode,
    ConnectionError,
    HttpHeaders,
    HttpxAsyncHttpNode,
    HttpxHttpNode,
    RequestsHttpNode,
    Urllib3HttpNode,
    debug_logging,
)
from elastic_transport._compat import await_if_coro
from elastic_transport._node._base import DEFAULT_USER_AGENT

node_class = pytest.mark.parametrize(
    "node_class",
    [
        Urllib3HttpNode,
        RequestsHttpNode,
        AiohttpHttpNode,
        HttpxAsyncHttpNode,
        HttpxHttpNode,
    ],
)


@node_class
@pytest.mark.anyio
async def test_debug_logging(node_class, anyio_backend, httpbin_node_config, httpbin):
    if anyio_backend == "trio" and node_class is not HttpxAsyncHttpNode:
        pytest.skip("only httpx supports trio")

    debug_logging()

    stream = io.StringIO()
    logging.getLogger("elastic_transport.node").addHandler(
        logging.StreamHandler(stream)
    )

    node = node_class(httpbin_node_config)
    await await_if_coro(
        node.perform_request(
            "GET",
            "/anything",
            body=b'{"key":"value"}',
            headers=HttpHeaders({"Content-Type": "application/json"}),
        )
    )

    print(node_class)
    print(stream.getvalue())

    response = stream.getvalue()
    print(response)
    for line in [
        "> GET /anything HTTP/1.1",
        "> Connection: keep-alive",
        "> Content-Type: application/json",
        f"> User-Agent: {DEFAULT_USER_AGENT}",
        '> {"key":"value"}',
        "< HTTP/1.1 200 OK",
        "< Access-Control-Allow-Credentials: true",
        "< Access-Control-Allow-Origin: *",
        "< Content-Type: application/json",
        "< {",
        '"args":{},',
        '"data":"{\\"key\\":\\"value\\"}",',
        '"files":{},',
        '"form":{},',
        '"headers":{',
        '"Content-Type":"application/json",',
        f'"Host":"{httpbin.host}:{httpbin.port}",',
        f'"User-Agent":"{DEFAULT_USER_AGENT}"',
        "},",
        '"json":{',
        '"key":"value"',
        "},",
        '"method":"GET",',
        f'"url":"{httpbin.url}/anything"',
        "}",
    ]:
        assert line in response


@node_class
@pytest.mark.anyio
async def test_debug_logging_uncompressed_body(
    httpbin_node_config, node_class, anyio_backend
):
    if anyio_backend == "trio" and node_class is not HttpxAsyncHttpNode:
        pytest.skip("only httpx supports trio")

    debug_logging()
    stream = io.StringIO()
    logging.getLogger("elastic_transport.node").addHandler(
        logging.StreamHandler(stream)
    )

    node = node_class(httpbin_node_config.replace(http_compress=True))
    await await_if_coro(
        node.perform_request(
            "GET",
            "/anything",
            body=b'{"key":"value"}',
            headers=HttpHeaders({"Content-Type": "application/json"}),
        )
    )

    lines = stream.getvalue().split("\n")
    print(lines)
    assert '> {"key":"value"}' in lines


@node_class
@pytest.mark.anyio
async def test_debug_logging_no_body(httpbin_node_config, node_class, anyio_backend):
    if anyio_backend == "trio" and node_class is not HttpxAsyncHttpNode:
        pytest.skip("only httpx supports trio")

    debug_logging()
    stream = io.StringIO()
    logging.getLogger("elastic_transport.node").addHandler(
        logging.StreamHandler(stream)
    )

    node = node_class(httpbin_node_config)
    await await_if_coro(
        node.perform_request(
            "HEAD",
            "/anything",
        )
    )

    lines = stream.getvalue().split("\n")[:-3]
    assert "> HEAD /anything HTTP/1.1" in lines


@node_class
@pytest.mark.anyio
async def test_debug_logging_error(httpbin_node_config, node_class, anyio_backend):
    if anyio_backend == "trio" and node_class is not HttpxAsyncHttpNode:
        pytest.skip("only httpx supports trio")

    debug_logging()
    stream = io.StringIO()
    logging.getLogger("elastic_transport.node").addHandler(
        logging.StreamHandler(stream)
    )

    node = node_class(httpbin_node_config.replace(host="not.a.valid.host"))
    with pytest.raises(ConnectionError):
        await await_if_coro(
            node.perform_request(
                "HEAD",
                "/anything",
            )
        )

    lines = stream.getvalue().split("\n")[:-3]
    assert "> HEAD /anything HTTP/?.?" in lines
    assert all(not line.startswith("<") for line in lines)



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import gzip
import json
import warnings

import aiohttp
import pytest
from multidict import CIMultiDict

from elastic_transport import AiohttpHttpNode, NodeConfig
from elastic_transport._node._base import DEFAULT_USER_AGENT


class TestAiohttpHttpNode:
    @pytest.mark.asyncio
    async def _get_mock_node(self, node_config, response_body=b"{}"):
        node = AiohttpHttpNode(node_config)
        node._create_aiohttp_session()

        def _dummy_request(*args, **kwargs):
            class DummyResponse:
                async def __aenter__(self, *_, **__):
                    return self

                async def __aexit__(self, *_, **__):
                    pass

                async def read(self):
                    return response_body if args[0] != "HEAD" else b""

                async def release(self):
                    return None

            dummy_response = DummyResponse()
            dummy_response.headers = CIMultiDict()
            dummy_response.status = 200
            _dummy_request.call_args = (args, kwargs)
            return dummy_response

        node.session.request = _dummy_request
        return node

    @pytest.mark.asyncio
    async def test_aiohttp_options(self):
        node = await self._get_mock_node(
            NodeConfig(scheme="http", host="localhost", port=80)
        )
        await node.perform_request(
            "GET",
            "/index",
            body=b"hello, world!",
            headers={"key": "value"},
        )

        args, kwargs = node.session.request.call_args
        assert args == ("GET", "http://localhost:80/index")
        assert kwargs == {
            "data": b"hello, world!",
            "headers": {
                "connection": "keep-alive",
                "key": "value",
                "user-agent": DEFAULT_USER_AGENT,
            },
            "timeout": aiohttp.ClientTimeout(
                total=10,
                connect=None,
                sock_read=None,
                sock_connect=None,
            ),
        }

    @pytest.mark.asyncio
    async def test_aiohttp_options_fingerprint(self):
        node = await self._get_mock_node(
            NodeConfig(
                scheme="https",
                host="localhost",
                port=443,
                ssl_assert_fingerprint=("00:" * 32).strip(":"),
            )
        )
        await node.perform_request(
            "GET",
            "/",
        )

        args, kwargs = node.session.request.call_args
        assert args == ("GET", "https://localhost:443/")

        # aiohttp.Fingerprint() doesn't define equality
        fingerprint: aiohttp.Fingerprint = kwargs.pop("ssl")
        assert fingerprint.fingerprint == b"\x00" * 32

        assert kwargs == {
            "data": None,
            "headers": {"connection": "keep-alive", "user-agent": DEFAULT_USER_AGENT},
            "timeout": aiohttp.ClientTimeout(
                total=10,
                connect=None,
                sock_read=None,
                sock_connect=None,
            ),
        }

    @pytest.mark.parametrize(
        "options",
        [(5, 5, 5), (None, 5, 5), (5, None, 0), (None, None, 0), (5, 5), (None, 0)],
    )
    @pytest.mark.asyncio
    async def test_aiohttp_options_timeout(self, options):
        if len(options) == 3:
            constructor_timeout, request_timeout, aiohttp_timeout = options
            node = await self._get_mock_node(
                NodeConfig(
                    scheme="http",
                    host="localhost",
                    port=80,
                    request_timeout=constructor_timeout,
                )
            )
        else:
            request_timeout, aiohttp_timeout = options
            node = await self._get_mock_node(
                NodeConfig(scheme="http", host="localhost", port=80)
            )

        await node.perform_request(
            "GET",
            "/",
            request_timeout=request_timeout,
        )

        args, kwargs = node.session.request.call_args
        assert args == ("GET", "http://localhost:80/")
        assert kwargs == {
            "data": None,
            "headers": {"connection": "keep-alive", "user-agent": DEFAULT_USER_AGENT},
            "timeout": aiohttp.ClientTimeout(
                total=aiohttp_timeout,
                connect=None,
                sock_read=None,
                sock_connect=None,
            ),
        }

    @pytest.mark.asyncio
    async def test_http_compression(self):
        node = await self._get_mock_node(
            NodeConfig(scheme="http", host="localhost", port=80, http_compress=True)
        )

        # 'content-encoding' shouldn't be set at a session level.
        # Should be applied only if the request is sent with a body.
        assert "content-encoding" not in node.session.headers

        await node.perform_request("GET", "/", body=b"{}")

        args, kwargs = node.session.request.call_args
        assert kwargs["headers"] == {
            "accept-encoding": "gzip",
            "connection": "keep-alive",
            "content-encoding": "gzip",
            "user-agent": DEFAULT_USER_AGENT,
        }
        assert gzip.decompress(kwargs["data"]) == b"{}"

    @pytest.mark.parametrize("http_compress", [None, False])
    @pytest.mark.asyncio
    async def test_no_http_compression(self, http_compress):
        node = await self._get_mock_node(
            NodeConfig(
                scheme="http", host="localhost", port=80, http_compress=http_compress
            )
        )

        assert "content-encoding" not in node.session.headers

        await node.perform_request("GET", "/", body=b"{}")

        args, kwargs = node.session.request.call_args
        assert kwargs["headers"] == {
            "connection": "keep-alive",
            "user-agent": DEFAULT_USER_AGENT,
        }
        assert kwargs["data"] == b"{}"

    @pytest.mark.parametrize("path_prefix", ["url", "/url"])
    @pytest.mark.asyncio
    async def test_uses_https_if_verify_certs_is_off(self, path_prefix):
        with warnings.catch_warnings(record=True) as w:
            await self._get_mock_node(
                NodeConfig(
                    scheme="https",
                    host="localhost",
                    port=443,
                    path_prefix=path_prefix,
                    verify_certs=False,
                )
            )

        assert 1 == len(w)
        assert (
            "Connecting to 'https://localhost:443/url' using TLS with verify_certs=False is insecure"
            == str(w[0].message)
        )

    @pytest.mark.asyncio
    async def test_uses_https_if_verify_certs_is_off_no_show_warning(self):
        with warnings.catch_warnings(record=True) as w:
            node = await self._get_mock_node(
                NodeConfig(
                    scheme="https",
                    host="localhost",
                    port=443,
                    path_prefix="url",
                    ssl_show_warn=False,
                )
            )
            await node.perform_request("GET", "/")

        assert w == []

    @pytest.mark.asyncio
    async def test_merge_headers(self):
        node = await self._get_mock_node(
            NodeConfig(
                scheme="https",
                host="localhost",
                port=443,
                headers={"h1": "v1", "h2": "v2"},
            )
        )
        resp, _ = await node.perform_request(
            "GET", "/", headers={"H2": "v2p", "H3": "v3"}
        )

        args, kwargs = node.session.request.call_args
        assert args == ("GET", "https://localhost:443/")
        assert kwargs["headers"] == {
            "connection": "keep-alive",
            "h1": "v1",
            "h2": "v2p",
            "h3": "v3",
            "user-agent": DEFAULT_USER_AGENT,
        }

    @pytest.mark.parametrize("aiohttp_fixed_head_bug", [True, False])
    @pytest.mark.asyncio
    async def test_head_workaround(self, aiohttp_fixed_head_bug):
        from elastic_transport._node import _http_aiohttp

        prev = _http_aiohttp._AIOHTTP_FIXED_HEAD_BUG
        try:
            _http_aiohttp._AIOHTTP_FIXED_HEAD_BUG = aiohttp_fixed_head_bug

            node = await self._get_mock_node(
                NodeConfig(
                    scheme="https",
                    host="localhost",
                    port=443,
                )
            )
            resp, data = await node.perform_request("HEAD", "/anything")

            method, url = node.session.request.call_args[0]
            assert method == "HEAD" if aiohttp_fixed_head_bug else "GET"
            assert url == "https://localhost:443/anything"

            assert resp.status == 200
            assert data == b""

        finally:
            _http_aiohttp._AIOHTTP_FIXED_HEAD_BUG = prev


@pytest.mark.asyncio
async def test_ssl_assert_fingerprint(cert_fingerprint, httpbin_secure):
    with warnings.catch_warnings(record=True) as w:
        node = AiohttpHttpNode(
            NodeConfig(
                scheme="https",
                host=httpbin_secure.host,
                port=httpbin_secure.port,
                ssl_assert_fingerprint=cert_fingerprint,
            )
        )
        resp, _ = await node.perform_request("GET", "/")

    assert resp.status == 200
    assert [str(x.message) for x in w if x.category != DeprecationWarning] == []


@pytest.mark.asyncio
async def test_default_headers(httpbin):
    node = AiohttpHttpNode(
        NodeConfig(scheme="http", host=httpbin.host, port=httpbin.port)
    )
    resp, data = await node.perform_request("GET", "/anything")

    assert resp.status == 200
    headers = json.loads(data)["headers"]
    headers.pop("X-Amzn-Trace-Id", None)
    assert headers == {
        "Connection": "keep-alive",
        "Host": f"{httpbin.host}:{httpbin.port}",
        "User-Agent": DEFAULT_USER_AGENT,
    }


@pytest.mark.asyncio
async def test_custom_headers(httpbin):
    node = AiohttpHttpNode(
        NodeConfig(
            scheme="http",
            host=httpbin.host,
            port=httpbin.port,
            headers={"accept-encoding": "gzip", "Content-Type": "application/json"},
        )
    )
    resp, data = await node.perform_request(
        "GET",
        "/anything",
        headers={
            "conTent-type": "application/x-ndjson",
            "user-agent": "custom-agent/1.2.3",
        },
    )

    assert resp.status == 200
    headers = json.loads(data)["headers"]
    headers.pop("X-Amzn-Trace-Id", None)
    assert headers == {
        "Accept-Encoding": "gzip",
        "Connection": "keep-alive",
        "Content-Type": "application/x-ndjson",
        "Host": f"{httpbin.host}:{httpbin.port}",
        "User-Agent": "custom-agent/1.2.3",
    }


@pytest.mark.asyncio
async def test_custom_user_agent(httpbin):
    node = AiohttpHttpNode(
        NodeConfig(
            scheme="http",
            host=httpbin.host,
            port=httpbin.port,
            headers={
                "accept-encoding": "gzip",
                "Content-Type": "application/json",
                "user-agent": "custom-agent/1.2.3",
            },
        )
    )
    resp, data = await node.perform_request(
        "GET",
        "/anything",
    )

    assert resp.status == 200
    headers = json.loads(data)["headers"]
    headers.pop("X-Amzn-Trace-Id", None)
    assert headers == {
        "Accept-Encoding": "gzip",
        "Connection": "keep-alive",
        "Content-Type": "application/json",
        "Host": f"{httpbin.host}:{httpbin.port}",
        "User-Agent": "custom-agent/1.2.3",
    }


def test_repr():
    node = AiohttpHttpNode(NodeConfig(scheme="https", host="localhost", port=443))
    assert "<AiohttpHttpNode(https://localhost:443)>" == repr(node)


@pytest.mark.asyncio
async def test_head(httpbin):
    node = AiohttpHttpNode(
        NodeConfig(
            scheme="http", host=httpbin.host, port=httpbin.port, http_compress=True
        )
    )
    resp, data = await node.perform_request("HEAD", "/anything")

    assert resp.status == 200
    assert resp.headers["content-type"] == "application/json"
    assert data == b""



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.


import random
import threading
import time

import pytest

from elastic_transport import NodeConfig, NodePool, Urllib3HttpNode


def test_node_pool_repr():
    node_configs = [NodeConfig("http", "localhost", x) for x in range(5)]
    random.shuffle(node_configs)
    pool = NodePool(node_configs, node_class=Urllib3HttpNode)
    assert repr(pool) == "<NodePool>"


def test_node_pool_empty_error():
    with pytest.raises(ValueError) as e:
        NodePool([], node_class=Urllib3HttpNode)
    assert str(e.value) == "Must specify at least one NodeConfig"


def test_node_pool_duplicate_node_configs():
    node_config = NodeConfig("http", "localhost", 80)
    with pytest.raises(ValueError) as e:
        NodePool([node_config, node_config], node_class=Urllib3HttpNode)
    assert str(e.value) == "Cannot use duplicate NodeConfigs within a NodePool"


def test_node_pool_get():
    node_config = NodeConfig("http", "localhost", 80)
    pool = NodePool([node_config], node_class=Urllib3HttpNode)
    assert pool.get().config is node_config


def test_node_pool_remove_seed_node():
    node_config = NodeConfig("http", "localhost", 80)
    pool = NodePool([node_config], node_class=Urllib3HttpNode)

    pool.remove(node_config)  # Calling .remove() on a seed node is a no-op
    assert len(pool._removed_nodes) == 0


def test_node_pool_add_and_remove_non_seed_node():
    node_config1 = NodeConfig("http", "localhost", 80)
    node_config2 = NodeConfig("http", "localhost", 81)
    pool = NodePool([node_config1], node_class=Urllib3HttpNode)

    pool.add(node_config2)
    assert any(pool.get().config is node_config2 for _ in range(10))

    pool.remove(node_config2)
    assert len(pool._removed_nodes) == 1

    # We never return a 'removed' node
    assert all(pool.get().config is node_config1 for _ in range(10))

    # We add it back, now we should .get() the node again.
    pool.add(node_config2)
    assert any(pool.get().config is node_config2 for _ in range(10))


def test_added_node_is_used_first():
    node_config1 = NodeConfig("http", "localhost", 80)
    node_config2 = NodeConfig("http", "localhost", 81)
    pool = NodePool([node_config1], node_class=Urllib3HttpNode)

    node1 = pool.get()
    pool.mark_dead(node1)

    pool.add(node_config2)
    assert pool.get().config is node_config2


def test_round_robin_selector():
    node_configs = [NodeConfig("http", "localhost", x) for x in range(5)]
    random.shuffle(node_configs)
    pool = NodePool(
        node_configs, node_class=Urllib3HttpNode, node_selector_class="round_robin"
    )

    get_node_configs = [pool.get() for _ in node_configs]
    for node_config in get_node_configs:
        assert pool.get() is node_config


@pytest.mark.parametrize(
    "node_configs",
    [
        [NodeConfig("http", "localhost", 80)],
        [NodeConfig("http", "localhost", 80), NodeConfig("http", "localhost", 81)],
    ],
)
def test_all_dead_nodes_still_gets_node(node_configs):
    pool = NodePool(node_configs, node_class=Urllib3HttpNode)

    for _ in node_configs:
        pool.mark_dead(pool.get())
    assert len(pool._alive_nodes) == 0

    node = pool.get()
    assert node.config in node_configs
    assert len(pool._alive_nodes) < 2


def test_unknown_selector_class():
    with pytest.raises(ValueError) as e:
        NodePool(
            [NodeConfig("http", "localhost", 80)],
            node_class=Urllib3HttpNode,
            node_selector_class="unknown",
        )
    assert str(e.value) == (
        "Unknown option for selector_class: 'unknown'. "
        "Available options are: 'random', 'round_robin'"
    )


def test_disable_randomize_nodes():
    node_configs = [NodeConfig("http", "localhost", x) for x in range(100)]
    pool = NodePool(node_configs, node_class=Urllib3HttpNode, randomize_nodes=False)

    assert [pool.get().config for _ in node_configs] == node_configs


def test_nodes_randomized_by_default():
    node_configs = [NodeConfig("http", "localhost", x) for x in range(100)]
    pool = NodePool(node_configs, node_class=Urllib3HttpNode)

    assert [pool.get().config for _ in node_configs] != node_configs


def test_dead_nodes_are_skipped():
    node_configs = [NodeConfig("http", "localhost", x) for x in range(2)]
    pool = NodePool(node_configs, node_class=Urllib3HttpNode)
    dead_node = pool.get()
    pool.mark_dead(dead_node)

    alive_node = pool.get()
    assert dead_node.config != alive_node.config

    assert all([pool.get().config == alive_node.config for _ in range(10)])


def test_dead_node_backoff_calculation():
    node_configs = [NodeConfig("http", "localhost", 9200)]
    pool = NodePool(
        node_configs,
        node_class=Urllib3HttpNode,
        dead_node_backoff_factor=0.5,
        max_dead_node_backoff=3.5,
    )
    node = pool.get()
    pool.mark_dead(node, _now=0)

    assert pool._dead_consecutive_failures == {node.config: 1}
    assert pool._dead_nodes.queue == [(0.5, node)]

    assert pool.get() is node
    pool.mark_dead(node, _now=0)

    assert pool._dead_consecutive_failures == {node.config: 2}
    assert pool._dead_nodes.queue == [(1.0, node)]

    assert pool.get() is node
    pool.mark_dead(node, _now=0)

    assert pool._dead_consecutive_failures == {node.config: 3}
    assert pool._dead_nodes.queue == [(2.0, node)]

    assert pool.get() is node
    pool.mark_dead(node, _now=0)

    assert pool._dead_consecutive_failures == {node.config: 4}
    assert pool._dead_nodes.queue == [(3.5, node)]

    assert pool.get() is node
    pool.mark_dead(node, _now=0)

    pool._dead_consecutive_failures = {node.config: 13292}
    assert pool._dead_nodes.queue == [(3.5, node)]

    assert pool.get() is node
    pool.mark_live(node)

    assert pool._dead_consecutive_failures == {}
    assert pool._dead_nodes.queue == []


def test_add_node_after_sniffing():
    node_configs = [NodeConfig("http", "localhost", 9200)]
    pool = NodePool(
        node_configs,
        node_class=Urllib3HttpNode,
    )

    # Initial node is marked as dead
    node = pool.get()
    pool.mark_dead(node)

    new_node_config = NodeConfig("http", "localhost", 9201)
    pool.add(new_node_config)

    # Internal flag is updated properly
    assert pool._all_nodes_len_1 is False

    # We get the new node instead of the old one
    new_node = pool.get()
    assert new_node.config == new_node_config

    # The old node is still on timeout so we should only get the new one.
    for _ in range(10):
        assert pool.get() is new_node


@pytest.mark.parametrize("pool_size", [1, 8])
def test_threading_test(pool_size):
    pool = NodePool(
        [NodeConfig("http", "localhost", x) for x in range(pool_size)],
        node_class=Urllib3HttpNode,
    )
    start = time.time()

    class ThreadTest(threading.Thread):
        def __init__(self):
            super().__init__()
            self.nodes_gotten = 0

        def run(self) -> None:
            while time.time() < start + 2:
                node = pool.get()
                self.nodes_gotten += 1
                if random.random() > 0.9:
                    pool.mark_dead(node)
                else:
                    pool.mark_live(node)

    threads = [ThreadTest() for _ in range(pool_size * 2)]
    [thread.start() for thread in threads]
    [thread.join() for thread in threads]

    assert sum(thread.nodes_gotten for thread in threads) >= 10000



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import nox

SOURCE_FILES = (
    "noxfile.py",
    "setup.py",
    "elastic_transport/",
    "utils/",
    "tests/",
    "docs/sphinx/",
)


@nox.session()
def format(session):
    session.install("black~=24.0", "isort", "pyupgrade")
    session.run("black", "--target-version=py310", *SOURCE_FILES)
    session.run("isort", *SOURCE_FILES)
    session.run("python", "utils/license-headers.py", "fix", *SOURCE_FILES)

    lint(session)


@nox.session
def lint(session):
    session.install(
        "flake8",
        "black~=24.0",
        "isort",
        "mypy==1.7.1",
        "types-requests",
        "types-certifi",
    )
    # https://github.com/python/typeshed/issues/10786
    session.run(
        "python", "-m", "pip", "uninstall", "--yes", "types-urllib3", silent=True
    )
    session.install(".[develop]")
    session.run("black", "--check", "--target-version=py310", *SOURCE_FILES)
    session.run("isort", "--check", *SOURCE_FILES)
    session.run("flake8", "--ignore=E501,W503,E203,E704", *SOURCE_FILES)
    session.run("python", "utils/license-headers.py", "check", *SOURCE_FILES)
    session.run("mypy", "--strict", "--show-error-codes", "elastic_transport/")


@nox.session(python=["3.10", "3.11", "3.12", "3.13", "3.14"])
def test(session):
    session.install(".[develop]")
    session.run(
        "pytest",
        "--cov=elastic_transport",
        *(session.posargs or ("tests/",)),
        env={"PYTHONWARNINGS": "always::DeprecationWarning"},
    )
    session.run("coverage", "report", "-m")


@nox.session(name="test-min-deps", python="3.10")
def test_min_deps(session):
    session.install("-r", "requirements-min.txt", ".[develop]", silent=False)
    session.run(
        "pytest",
        "--cov=elastic_transport",
        *(session.posargs or ("tests/",)),
        env={"PYTHONWARNINGS": "always::DeprecationWarning"},
    )
    session.run("coverage", "report", "-m")


@nox.session(python="3")
def docs(session):
    session.install(".[develop]")

    session.chdir("docs/sphinx")
    session.run(
        "sphinx-build",
        "-T",
        "-E",
        "-b",
        "html",
        "-d",
        "_build/doctrees",
        "-D",
        "language=en",
        ".",
        "_build/html",
    )



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

"""A command line tool for building and verifying releases
Can be used for building both 'elasticsearch' and 'elasticsearchX' dists.
Only requires 'name' in 'setup.py' and the directory to be changed.
"""

import contextlib
import os
import re
import shutil
import tempfile

base_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
tmp_dir = None


def shlex_quote(s):
    # Backport of shlex.quote() to Python 2.x
    _find_unsafe = re.compile(r"[^\w@%+=:,./-]").search

    if not s:
        return "''"
    if _find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'
    return "'" + s.replace("'", "'\"'\"'") + "'"


@contextlib.contextmanager
def set_tmp_dir():
    global tmp_dir
    tmp_dir = tempfile.mkdtemp()
    yield tmp_dir
    shutil.rmtree(tmp_dir)
    tmp_dir = None


def run(argv, expect_exit_code=0):
    if tmp_dir is None:
        os.chdir(base_dir)
    else:
        os.chdir(tmp_dir)

    cmd = " ".join(shlex_quote(x) for x in argv)
    print("$ " + cmd)
    exit_code = os.system(cmd)
    if exit_code != expect_exit_code:
        print(
            "Command exited incorrectly: should have been %d was %d"
            % (expect_exit_code, exit_code)
        )
        exit(exit_code or 1)


def test_dist(dist):
    with set_tmp_dir() as tmp_dir:
        # Build the venv and install the dist
        run(("python", "-m", "venv", os.path.join(tmp_dir, "venv")))
        venv_python = os.path.join(tmp_dir, "venv/bin/python")
        run((venv_python, "-m", "pip", "install", "-U", "pip"))
        run((venv_python, "-m", "pip", "install", dist))

        # Test out importing from the package
        run(
            (
                venv_python,
                "-c",
                "from elastic_transport import Transport, Urllib3HttpNode, RequestsHttpNode",
            )
        )

        # Uninstall the dist, see that we can't import things anymore
        run((venv_python, "-m", "pip", "uninstall", "--yes", "elastic-transport"))
        run(
            (venv_python, "-c", "from elastic_transport import Transport"),
            expect_exit_code=256,
        )


def main():
    run(("rm", "-rf", "build/", "dist/", "*.egg-info", ".eggs"))

    # Install and run python-build to create sdist/wheel
    run(("python", "-m", "pip", "install", "-U", "build"))
    run(("python", "-m", "build"))

    for dist in os.listdir(os.path.join(base_dir, "dist")):
        test_dist(os.path.join(base_dir, "dist", dist))

    # After this run 'python -m twine upload dist/*'
    print(
        "\n\n"
        "===============================\n\n"
        "    * Releases are ready! *\n\n"
        "$ python -m twine upload dist/*\n\n"
        "==============================="
    )


if __name__ == "__main__":
    main()



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import datetime
import os
import sys

sys.path.insert(0, os.path.abspath("../.."))

from elastic_transport import __version__  # noqa

project = "elastic-transport"
copyright = f"{datetime.date.today().year} Elasticsearch B.V."
author = "Seth Michael Larson"
version = __version__
release = __version__

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx_autodoc_typehints",
]

pygments_style = "sphinx"
pygments_dark_style = "monokai"

templates_path = []
exclude_patterns = []
html_theme = "furo"
html_static_path = []

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "requests": ("https://requests.readthedocs.io/en/latest", None),
}



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

from platform import python_version

import pytest

from elastic_transport import Urllib3HttpNode, __version__
from elastic_transport.client_utils import (
    basic_auth_to_header,
    client_meta_version,
    create_user_agent,
    parse_cloud_id,
    url_to_node_config,
)


def test_create_user_agent():
    assert create_user_agent(
        "enterprise-search-python", "7.10.0"
    ) == "enterprise-search-python/7.10.0 (Python/{}; elastic-transport/{})".format(
        python_version(),
        __version__,
    )


@pytest.mark.parametrize(
    ["version", "meta_version"],
    [
        ("7.10.0", "7.10.0"),
        ("7.10.0-alpha1", "7.10.0p"),
        ("3.9.0b1", "3.9.0p"),
        ("3.9.pre1", "3.9p"),
        ("3.7.4.post1", "3.7.4"),
        ("3.7.4.post", "3.7.4"),
    ],
)
def test_client_meta_version(version, meta_version):
    assert client_meta_version(version) == meta_version


def test_parse_cloud_id():
    cloud_id = parse_cloud_id(
        "cluster:dXMtZWFzdC0xLmF3cy5mb3VuZC5pbyQ0ZmE4ODIxZTc1NjM0MDMyYmVk"
        "MWNmMjIxMTBlMmY5NyQ0ZmE4ODIxZTc1NjM0MDMyYmVkMWNmMjIxMTBlMmY5Ng=="
    )
    assert cloud_id.cluster_name == "cluster"
    assert cloud_id.es_address == (
        "4fa8821e75634032bed1cf22110e2f97.us-east-1.aws.found.io",
        443,
    )
    assert cloud_id.kibana_address == (
        "4fa8821e75634032bed1cf22110e2f96.us-east-1.aws.found.io",
        443,
    )


@pytest.mark.parametrize(
    ["cloud_id", "port"],
    [
        (
            ":dXMtZWFzdC0xLmF3cy5mb3VuZC5pbzo5MjQzJDRmYTg4MjFlNzU2MzQwMzJiZ"
            "WQxY2YyMjExMGUyZjk3JDRmYTg4MjFlNzU2MzQwMzJiZWQxY2YyMjExMGUyZjk2",
            9243,
        ),
        (
            ":dXMtZWFzdC0xLmF3cy5mb3VuZC5pbzo0NDMkNGZhODgyMWU3NTYzNDAzMmJlZD"
            "FjZjIyMTEwZTJmOTckNGZhODgyMWU3NTYzNDAzMmJlZDFjZjIyMTEwZTJmOTY=",
            443,
        ),
    ],
)
def test_parse_cloud_id_ports(cloud_id, port):
    cloud_id = parse_cloud_id(cloud_id)
    assert cloud_id.cluster_name == ""
    assert cloud_id.es_address == (
        "4fa8821e75634032bed1cf22110e2f97.us-east-1.aws.found.io",
        port,
    )
    assert cloud_id.kibana_address == (
        "4fa8821e75634032bed1cf22110e2f96.us-east-1.aws.found.io",
        port,
    )


@pytest.mark.parametrize(
    "cloud_id",
    [
        "cluster:dXMtZWFzdC0xLmF3cy5mb3VuZC5pbyQ0ZmE4ODIxZTc1NjM0MDMyYmVkMWNmMjIxMTBlMmY5NyQ=",
        "cluster:dXMtZWFzdC0xLmF3cy5mb3VuZC5pbyQ0ZmE4ODIxZTc1NjM0MDMyYmVkMWNmMjIxMTBlMmY5Nw==",
    ],
)
def test_parse_cloud_id_no_kibana(cloud_id):
    cloud_id = parse_cloud_id(cloud_id)
    assert cloud_id.cluster_name == "cluster"
    assert cloud_id.es_address == (
        "4fa8821e75634032bed1cf22110e2f97.us-east-1.aws.found.io",
        443,
    )
    assert cloud_id.kibana_address is None


@pytest.mark.parametrize(
    "cloud_id",
    [
        "cluster:dXMtZWFzdC0xLmF3cy5mb3VuZC5pbzo0NDMkJA==",
        "cluster:dXMtZWFzdC0xLmF3cy5mb3VuZC5pbzo0NDM=",
    ],
)
def test_parse_cloud_id_no_es(cloud_id):
    cloud_id = parse_cloud_id(cloud_id)
    assert cloud_id.cluster_name == "cluster"
    assert cloud_id.es_address is None
    assert cloud_id.kibana_address is None


@pytest.mark.parametrize(
    "cloud_id",
    [
        "cluster:",
        "dXMtZWFzdC0xLmF3cy5mb3VuZC5pbyQ0ZmE4ODIxZTc1NjM0MDMyYmVkMWNmMjIxMTBlMmY5NyQ=",
        "cluster:ā",
    ],
)
def test_invalid_cloud_id(cloud_id):
    with pytest.raises(ValueError) as e:
        parse_cloud_id(cloud_id)
    assert str(e.value) == "Cloud ID is not properly formatted"


@pytest.mark.parametrize(
    ["url", "node_base_url", "path_prefix"],
    [
        ("https://localhost", "https://localhost:443", ""),
        ("http://localhost:3002", "http://localhost:3002", ""),
        ("http://127.0.0.1:3002", "http://127.0.0.1:3002", ""),
        ("http://127.0.0.1:3002/", "http://127.0.0.1:3002", ""),
        (
            "http://127.0.0.1:3002/path-prefix",
            "http://127.0.0.1:3002/path-prefix",
            "/path-prefix",
        ),
        (
            "http://localhost:3002/url-prefix/",
            "http://localhost:3002/url-prefix",
            "/url-prefix",
        ),
        (
            "https://localhost/url-prefix",
            "https://localhost:443/url-prefix",
            "/url-prefix",
        ),
        ("http://[::1]:3002/url-prefix", "http://[::1]:3002/url-prefix", "/url-prefix"),
        ("https://[::1]:0/", "https://[::1]:0", ""),
    ],
)
def test_url_to_node_config(url, node_base_url, path_prefix):
    node_config = url_to_node_config(url)
    assert Urllib3HttpNode(node_config).base_url == node_base_url

    assert "[" not in node_config.host
    assert isinstance(node_config.port, int)
    assert node_config.path_prefix == path_prefix
    assert url.lower().startswith(node_config.scheme)


@pytest.mark.parametrize(
    "url",
    [
        "localhost:0",
        "[::1]:3002/url-prefix",
        "localhost",
        "localhost/",
        "localhost:3",
        "[::1]/url-prefix/",
        "[::1]",
        "[::1]:3002",
        "http://localhost",
        "localhost/url-prefix/",
        "localhost:3002/url-prefix",
        "http://localhost/url-prefix",
    ],
)
def test_url_to_node_config_error_missing_component(url):
    with pytest.raises(ValueError) as e:
        url_to_node_config(url)
    assert (
        str(e.value)
        == "URL must include a 'scheme', 'host', and 'port' component (ie 'https://localhost:9200')"
    )


@pytest.mark.parametrize(
    ["url", "port"],
    [
        ("http://127.0.0.1", 80),
        ("http://[::1]", 80),
        ("HTTPS://localhost", 443),
        ("https://localhost/url-prefix", 443),
    ],
)
def test_url_to_node_config_use_default_ports_for_scheme(url, port):
    node_config = url_to_node_config(url, use_default_ports_for_scheme=True)
    assert node_config.port == port


def test_url_with_auth_into_authorization():
    node_config = url_to_node_config("http://localhost:9200")
    assert node_config.headers == {}

    node_config = url_to_node_config("http://@localhost:9200")
    assert node_config.headers == {}

    node_config = url_to_node_config("http://user:pass@localhost:9200")
    assert node_config.headers == {"Authorization": "Basic dXNlcjpwYXNz"}

    node_config = url_to_node_config("http://user:@localhost:9200")
    assert node_config.headers == {"Authorization": "Basic dXNlcjo="}

    node_config = url_to_node_config("http://user@localhost:9200")
    assert node_config.headers == {"Authorization": "Basic dXNlcjo="}

    node_config = url_to_node_config("http://me@example.com:password@localhost:9200")
    assert node_config.headers == {
        "Authorization": "Basic bWVAZXhhbXBsZS5jb206cGFzc3dvcmQ="
    }

    # ensure username and password are passed to basic auth unmodified
    basic_auth = basic_auth_to_header(("user:@", "@password"))
    node_config = url_to_node_config("http://user:@:@password@localhost:9200")
    assert node_config.headers == {"Authorization": basic_auth}
    node_config = url_to_node_config("http://user%3A%40:%40password@localhost:9200")
    assert node_config.headers == {"Authorization": basic_auth}


@pytest.mark.parametrize(
    "basic_auth", ["", b"", ("",), ("", 1), (1, ""), ["", ""], False, object()]
)
def test_basic_auth_errors(basic_auth):
    with pytest.raises(ValueError) as e:
        basic_auth_to_header(basic_auth)
    assert (
        str(e.value)
        == "'basic_auth' must be a 2-tuple of str/bytes (username, password)"
    )



#  Licensed to Elasticsearch B.V. under one or more contributor
#  license agreements. See the NOTICE file distributed with
#  this work for additional information regarding copyright
#  ownership. Elasticsearch B.V. licenses this file to you under
#  the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing,
#  software distributed under the License is distributed on an
#  "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
#  KIND, either express or implied.  See the License for the
#  specific language governing permissions and limitations
#  under the License.

import gzip
import ssl
import warnings

import pytest
import respx

from elastic_transport import HttpxAsyncHttpNode, HttpxHttpNode, NodeConfig
from elastic_transport._node._base import DEFAULT_USER_AGENT


def create_sync_node(node_config: NodeConfig):
    return HttpxHttpNode(node_config)


class TestHttpxNodeCreation:
    def test_ssl_context(self):
        ssl_context = ssl.create_default_context()
        with warnings.catch_warnings(record=True) as w:
            node = create_sync_node(
                NodeConfig(
                    scheme="https",
                    host="localhost",
                    port=80,
                    ssl_context=ssl_context,
                )
            )
        assert node.client._transport._pool._ssl_context is ssl_context
        assert len(w) == 0

    def test_uses_https_if_verify_certs_is_off(self):
        with warnings.catch_warnings(record=True) as w:
            _ = create_sync_node(
                NodeConfig("https", "localhost", 443, verify_certs=False)
            )
        assert (
            str(w[0].message)
            == "Connecting to 'https://localhost:443' using TLS with verify_certs=False is insecure"
        )

    def test_no_warn_when_uses_https_if_verify_certs_is_off(self):
        with warnings.catch_warnings(record=True) as w:
            _ = create_sync_node(
                NodeConfig(
                    "https",
                    "localhost",
                    443,
                    verify_certs=False,
                    ssl_show_warn=False,
                )
            )
        assert 0 == len(w)

    def test_ca_certs_with_verify_ssl_false_raises_error(self):
        with pytest.raises(ValueError) as exc:
            create_sync_node(
                NodeConfig(
                    "https",
                    "localhost",
                    443,
                    ca_certs="/ca/certs",
                    verify_certs=False,
                )
            )
            assert (
                str(exc.value) == "You cannot use 'ca_certs' when 'verify_certs=False'"
            )

    def test_path_prefix(self):
        node = create_sync_node(
            NodeConfig(
                "http",
                "localhost",
                9200,
                path_prefix="/test",
            )
        )
        assert node.base_url == "http://localhost:9200/test"
        assert node.client.base_url == "http://localhost:9200/test/"


class TestHttpxNode:
    @respx.mock
    def test_simple_request(self):
        node = create_sync_node(NodeConfig(scheme="http", host="localhost", port=80))
        respx.get("http://localhost/index")
        node.perform_request("GET", "/index", b"hello world", headers={"key": "value"})
        request = respx.calls.last.request
        assert request.content == b"hello world"
        assert {
            "key": "value",
            "connection": "keep-alive",
            "user-agent": DEFAULT_USER_AGENT,
        }.items() <= request.headers.items()

    @respx.mock
    def test_compression(self):
        node = create_sync_node(
            NodeConfig(scheme="http", host="localhost", port=80, http_compress=True)
        )
        respx.get("http://localhost/index")
        node.perform_request("GET", "/index", b"hello world")
        request = respx.calls.last.request
        assert gzip.decompress(request.content) == b"hello world"
        assert {"content-encoding": "gzip"}.items() <= request.headers.items()

    @respx.mock
    def test_default_timeout(self):
        node = create_sync_node(
            NodeConfig(scheme="http", host="localhost", port=80, request_timeout=10)
        )
        respx.get("http://localhost/index")
        node.perform_request("GET", "/index", b"hello world")
        request = respx.calls.last.request
        assert request.extensions["timeout"]["connect"] == 10

    @respx.mock
    def test_overwritten_timeout(self):
        node = create_sync_node(
            NodeConfig(scheme="http", host="localhost", port=80, request_timeout=10)
        )
        respx.get("http://localhost/index")
        node.perform_request("GET", "/index", b"hello world", request_timeout=15)
        request = respx.calls.last.request
        assert request.extensions["timeout"]["connect"] == 15

    @respx.mock
    def test_merge_headers(self):
        node = create_sync_node(
            NodeConfig("http", "localhost", 80, headers={"h1": "v1", "h2": "v2"})
        )
        respx.get("http://localhost/index")
        node.perform_request(
            "GET", "/index", b"hello world", headers={"h2": "v2p", "h3": "v3"}
        )
        request = respx.calls.last.request
        assert request.headers["h1"] == "v1"
        assert request.headers["h2"] == "v2p"
        assert request.headers["h3"] == "v3"


def create_async_node(node_config: NodeConfig):
    return HttpxAsyncHttpNode(node_config)


class TestHttpxAsyncNodeCreation:
    def test_ssl_context(self):
        ssl_context = ssl.create_default_context()
        with warnings.catch_warnings(record=True) as w:
            node = create_async_node(
                NodeConfig(
                    scheme="https",
                    host="localhost",
                    port=80,
                    ssl_context=ssl_context,
                )
            )
        assert node.client._transport._pool._ssl_context is ssl_context
        assert len(w) == 0

    def test_uses_https_if_verify_certs_is_off(self):
        with warnings.catch_warnings(record=True) as w:
            _ = create_async_node(
                NodeConfig("https", "localhost", 443, verify_certs=False)
            )
        assert (
            str(w[0].message)
            == "Connecting to 'https://localhost:443' using TLS with verify_certs=False is insecure"
        )

    def test_no_warn_when_uses_https_if_verify_certs_is_off(self):
        with warnings.catch_warnings(record=True) as w:
            _ = create_async_node(
                NodeConfig(
                    "https",
                    "localhost",
                    443,
                    verify_certs=False,
                    ssl_show_warn=False,
                )
            )
        assert 0 == len(w)

    def test_ca_certs_with_verify_ssl_false_raises_error(self):
        with pytest.raises(ValueError) as exc:
            create_async_node(
                NodeConfig(
                    "https",
                    "localhost",
                    443,
                    ca_certs="/ca/certs",
                    verify_certs=False,
                )
            )
            assert (
                str(exc.value) == "You cannot use 'ca_certs' when 'verify_certs=False'"
            )

    def test_path_prefix(self):
        node = create_async_node(
            NodeConfig(
                "http",
                "localhost",
                9200,
                path_prefix="/test",
            )
        )
        assert node.base_url == "http://localhost:9200/test"
        assert node.client.base_url == "http://localhost:9200/test/"


@pytest.mark.anyio
class TestHttpxAsyncNode:
    @respx.mock
    async def test_simple_request(self):
        node = create_async_node(NodeConfig(scheme="http", host="localhost", port=80))
        respx.get("http://localhost/index")
        await node.perform_request(
            "GET", "/index", b"hello world", headers={"key": "value"}
        )
        request = respx.calls.last.request
        assert request.content == b"hello world"
        assert {
            "key": "value",
            "connection": "keep-alive",
            "user-agent": DEFAULT_USER_AGENT,
        }.items() <= request.headers.items()

    @respx.mock
    async def test_compression(self):
        node = create_async_node(
            NodeConfig(scheme="http", host="localhost", port=80, http_compress=True)
        )
        respx.get("http://localhost/index")
        await node.perform_request("GET", "/index", b"hello world")
        request = respx.calls.last.request
        assert gzip.decompress(request.content) == b"hello world"
        assert {"content-encoding": "gzip"}.items() <= request.headers.items()

    @respx.mock
    async def test_default_timeout(self):
        node = create_async_node(
            NodeConfig(scheme="http", host="localhost", port=80, request_timeout=10)
        )
        respx.get("http://localhost/index")
        await node.perform_request("GET", "/index", b"hello world")
        request = respx.calls.last.request
        assert request.extensions["timeout"]["connect"] == 10

    @respx.mock
    async def test_overwritten_timeout(self):
        node = create_async_node(
            NodeConfig(scheme="http", host="localhost", port=80, request_timeout=10)
        )
        respx.get("http://localhost/index")
        await node.perform_request("GET", "/index", b"hello world", request_timeout=15)
        request = respx.calls.last.request
        assert request.extensions["timeout"]["connect"] == 15

    @respx.mock
    async def test_merge_headers(self):
        node = create_async_node(
            NodeConfig("http", "localhost", 80, headers={"h1": "v1", "h2": "v2"})
        )
        respx.get("http://localhost/index")
        await node.perform_request(
            "GET", "/index", b"hello world", headers={"h2": "v2p", "h3": "v3"}
        )
        request = respx.calls.last.request
        assert request.headers["h1"] == "v1"
        assert request.headers["h2"] == "v2p"
        assert request.headers["h3"] == "v3"


@pytest.mark.parametrize("node_class", [HttpxHttpNode, HttpxAsyncHttpNode])
def test_ssl_assert_fingerprint(node_class, cert_fingerprint, httpbin_secure):
    with pytest.raises(ValueError, match="httpx does not support certificate pinning"):
        node_class(
            NodeConfig(
                scheme="https",
                host=httpbin_secure.host,
                port=httpbin_secure.port,
                ssl_assert_fingerprint=cert_fingerprint,
            )
        )
